# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Common graph operations for execution."""
import sys
from copy import deepcopy
from glob import glob
import os
import shutil
from time import sleep, time
from traceback import format_exception

import numpy as np

from ... import logging
from ...utils.misc import str2bool
from ..engine.utils import topological_sort, load_resultfile
from ..engine import MapNode
from .tools import report_crash, report_nodes_not_run, create_pyscript

logger = logging.getLogger("nipype.workflow")


class PluginBase(object):
    """Base class for plugins."""

    def __init__(self, plugin_args=None):
        if plugin_args is None:
            plugin_args = {}
        self.plugin_args = plugin_args
        self._config = None
        self._status_callback = plugin_args.get("status_callback")

    def run(self, graph, config, updatehash=False):
        """
        Instruct the plugin to execute the workflow graph.

        The core plugin member that should be implemented by
        all plugins.

        Parameters
        ----------
        graph :
            a networkx, flattened :abbr:`DAG (Directed Acyclic Graph)`
            to be executed
        config : :obj:`~nipype.config`
            a nipype.config object
        updatehash : :obj:`bool`
            whether cached nodes with stale hash should be just updated.

        """
        raise NotImplementedError


class DistributedPluginBase(PluginBase):
    """
    Execute workflow with a distribution engine

    Combinations of ``proc_done`` and ``proc_pending``:

    +------------+---------------+--------------------------------+
    | proc_done  | proc_pending  | outcome                        |
    +============+===============+================================+
    | True       | False         | Process is finished            |
    +------------+---------------+--------------------------------+
    | True       | True          | Process is currently being run |
    +------------+---------------+--------------------------------+
    | False      | False         | Process is queued              |
    +------------+---------------+--------------------------------+
    | False      | True          | INVALID COMBINATION            |
    +------------+---------------+--------------------------------+

    Attributes
    ----------
    procs : :obj:`list`
        list (N) of underlying interface elements to be processed
    proc_done : :obj:`numpy.ndarray`
        a boolean numpy array (N,) signifying whether a process has been
        submitted for execution
    proc_pending : :obj:`numpy.ndarray`
        a boolean numpy array (N,) signifying whether a
        process is currently running.
    depidx : :obj:`numpy.matrix`
        a boolean matrix (NxN) storing the dependency structure accross
        processes. Process dependencies are derived from each column.

    """

    def __init__(self, plugin_args=None):
        """
        Initialize runtime attributes to none

        """
        super(DistributedPluginBase, self).__init__(plugin_args=plugin_args)
        self.procs = None
        self.depidx = None
        self.refidx = None
        self.mapnodes = None
        self.mapnodesubids = None
        self.proc_done = None
        self.proc_pending = None
        self.pending_tasks = []
        self.max_jobs = self.plugin_args.get("max_jobs", np.inf)

    def _prerun_check(self, graph):
        """Stub method to validate/massage graph and nodes before running"""

    def _postrun_check(self):
        """Stub method to close any open resources"""

    def run(self, graph, config, updatehash=False):
        """
        Executes a pre-defined pipeline using distributed approaches
        """
        logger.info("Running in parallel.")
        self._config = config
        poll_sleep_secs = float(config["execution"]["poll_sleep_duration"])

        self._prerun_check(graph)
        # Generate appropriate structures for worker-manager model
        self._generate_dependency_list(graph)
        self.mapnodes = []
        self.mapnodesubids = {}
        # setup polling - TODO: change to threaded model
        notrun = []

        old_progress_stats = None
        old_presub_stats = None
        while not np.all(self.proc_done) or np.any(self.proc_pending):
            loop_start = time()
            # Check if a job is available (jobs with all dependencies run)
            # https://github.com/nipy/nipype/pull/2200#discussion_r141605722
            jobs_ready = np.nonzero(~self.proc_done & (self.depidx.sum(0) == 0))[1]

            progress_stats = (
                len(self.proc_done),
                np.sum(self.proc_done ^ self.proc_pending),
                np.sum(self.proc_done & self.proc_pending),
                len(jobs_ready),
                len(self.pending_tasks),
                np.sum(~self.proc_done & ~self.proc_pending),
            )
            display_stats = progress_stats != old_progress_stats
            if display_stats:
                logger.debug(
                    "Progress: %d jobs, %d/%d/%d "
                    "(done/running/ready), %d/%d "
                    "(pending_tasks/waiting).",
                    *progress_stats
                )
                old_progress_stats = progress_stats
            toappend = []
            # trigger callbacks for any pending results
            while self.pending_tasks:
                taskid, jobid = self.pending_tasks.pop()
                try:
                    result = self._get_result(taskid)
                except Exception:
                    notrun.append(self._clean_queue(jobid, graph))
                else:
                    if result:
                        if result["traceback"]:
                            notrun.append(
                                self._clean_queue(jobid, graph, result=result)
                            )
                        else:
                            self._task_finished_cb(jobid)
                            self._remove_node_dirs()
                        self._clear_task(taskid)
                    else:
                        assert self.proc_done[jobid] and self.proc_pending[jobid]
                        toappend.insert(0, (taskid, jobid))

            if toappend:
                self.pending_tasks.extend(toappend)

            num_jobs = len(self.pending_tasks)
            presub_stats = (num_jobs, np.sum(self.proc_done & self.proc_pending))
            display_stats = display_stats or presub_stats != old_presub_stats
            if display_stats:
                logger.debug("Tasks currently running: %d. Pending: %d.", *presub_stats)
                old_presub_stats = presub_stats
            if num_jobs < self.max_jobs:
                self._send_procs_to_workers(updatehash=updatehash, graph=graph)
            elif display_stats:
                logger.debug("Not submitting (max jobs reached)")

            sleep_til = loop_start + poll_sleep_secs
            sleep(max(0, sleep_til - time()))

        self._remove_node_dirs()
        report_nodes_not_run(notrun)

        # close any open resources
        self._postrun_check()

    def _get_result(self, taskid):
        raise NotImplementedError

    def _submit_job(self, node, updatehash=False):
        raise NotImplementedError

    def _report_crash(self, node, result=None):
        tb = None
        if result is not None:
            node._result = result["result"]
            tb = result["traceback"]
            node._traceback = tb
        return report_crash(node, traceback=tb)

    def _clear_task(self, taskid):
        raise NotImplementedError

    def _clean_queue(self, jobid, graph, result=None):
        logger.debug("Clearing %d from queue", jobid)

        if self._status_callback:
            self._status_callback(self.procs[jobid], "exception")
        if result is None:
            result = {
                "result": None,
                "traceback": "\n".join(format_exception(*sys.exc_info())),
            }

        crashfile = self._report_crash(self.procs[jobid], result=result)
        if str2bool(self._config["execution"]["stop_on_first_crash"]):
            raise RuntimeError("".join(result["traceback"]))
        if jobid in self.mapnodesubids:
            # remove current jobid
            self.proc_pending[jobid] = False
            self.proc_done[jobid] = True
            # remove parent mapnode
            jobid = self.mapnodesubids[jobid]
            self.proc_pending[jobid] = False
            self.proc_done[jobid] = True
        # remove dependencies from queue
        return self._remove_node_deps(jobid, crashfile, graph)

    def _submit_mapnode(self, jobid):
        import scipy.sparse as ssp

        if jobid in self.mapnodes:
            return True
        self.mapnodes.append(jobid)
        mapnodesubids = self.procs[jobid].get_subnodes()
        numnodes = len(mapnodesubids)
        logger.debug("Adding %d jobs for mapnode %s", numnodes, self.procs[jobid])
        for i in range(numnodes):
            self.mapnodesubids[self.depidx.shape[0] + i] = jobid
        self.procs.extend(mapnodesubids)
        self.depidx = ssp.vstack(
            (self.depidx, ssp.lil_matrix(np.zeros((numnodes, self.depidx.shape[1])))),
            "lil",
        )
        self.depidx = ssp.hstack(
            (self.depidx, ssp.lil_matrix(np.zeros((self.depidx.shape[0], numnodes)))),
            "lil",
        )
        self.depidx[-numnodes:, jobid] = 1
        self.proc_done = np.concatenate(
            (self.proc_done, np.zeros(numnodes, dtype=bool))
        )
        self.proc_pending = np.concatenate(
            (self.proc_pending, np.zeros(numnodes, dtype=bool))
        )
        return False

    def _send_procs_to_workers(self, updatehash=False, graph=None):
        """
        Sends jobs to workers
        """

        while not np.all(self.proc_done):
            num_jobs = len(self.pending_tasks)
            if np.isinf(self.max_jobs):
                slots = None
            else:
                slots = max(0, self.max_jobs - num_jobs)
            logger.debug("Slots available: %s", slots)
            if (num_jobs >= self.max_jobs) or (slots == 0):
                break

            # Check if a job is available (jobs with all dependencies run)
            # https://github.com/nipy/nipype/pull/2200#discussion_r141605722
            jobids = np.nonzero(~self.proc_done & (self.depidx.sum(0) == 0))[1]

            if len(jobids) > 0:
                # send all available jobs
                logger.info(
                    "Pending[%d] Submitting[%d] jobs Slots[%s]",
                    num_jobs,
                    len(jobids[:slots]),
                    slots or "inf",
                )

                for jobid in jobids[:slots]:
                    if isinstance(self.procs[jobid], MapNode):
                        try:
                            num_subnodes = self.procs[jobid].num_subnodes()
                        except Exception:
                            self._clean_queue(jobid, graph)
                            self.proc_pending[jobid] = False
                            continue
                        if num_subnodes > 1:
                            submit = self._submit_mapnode(jobid)
                            if not submit:
                                continue
                    # change job status in appropriate queues
                    self.proc_done[jobid] = True
                    self.proc_pending[jobid] = True
                    # Send job to task manager and add to pending tasks
                    logger.info("Submitting: %s ID: %d", self.procs[jobid], jobid)
                    if self._status_callback:
                        self._status_callback(self.procs[jobid], "start")

                    if not self._local_hash_check(jobid, graph):
                        if self.procs[jobid].run_without_submitting:
                            logger.debug(
                                "Running node %s on master thread", self.procs[jobid]
                            )
                            try:
                                self.procs[jobid].run()
                            except Exception:
                                self._clean_queue(jobid, graph)
                            self._task_finished_cb(jobid)
                            self._remove_node_dirs()
                        else:
                            tid = self._submit_job(
                                deepcopy(self.procs[jobid]), updatehash=updatehash
                            )
                            if tid is None:
                                self.proc_done[jobid] = False
                                self.proc_pending[jobid] = False
                            else:
                                self.pending_tasks.insert(0, (tid, jobid))
                    logger.info(
                        "Finished submitting: %s ID: %d", self.procs[jobid], jobid
                    )
            else:
                break

    def _local_hash_check(self, jobid, graph):
        if not str2bool(self.procs[jobid].config["execution"]["local_hash_check"]):
            return False

        try:
            cached, updated = self.procs[jobid].is_cached()
        except Exception:
            logger.warning(
                "Error while checking node hash, forcing re-run. "
                "Although this error may not prevent the workflow from running, "
                "it could indicate a major problem. Please report a new issue "
                "at https://github.com/nipy/nipype/issues adding the following "
                "information:\n\n\tNode: %s\n\tInterface: %s.%s\n\tTraceback:\n%s",
                self.procs[jobid],
                self.procs[jobid].interface.__module__,
                self.procs[jobid].interface.__class__.__name__,
                "\n".join(format_exception(*sys.exc_info())),
            )
            return False

        logger.debug(
            'Checking hash "%s" locally: cached=%s, updated=%s.',
            self.procs[jobid],
            cached,
            updated,
        )
        overwrite = self.procs[jobid].overwrite
        always_run = self.procs[jobid].interface.always_run

        if (
            cached
            and updated
            and (overwrite is False or overwrite is None and not always_run)
        ):
            logger.debug(
                "Skipping cached node %s with ID %s.", self.procs[jobid], jobid
            )
            try:
                self._task_finished_cb(jobid, cached=True)
                self._remove_node_dirs()
            except Exception:
                logger.debug(
                    "Error skipping cached node %s (%s).\n\n%s",
                    self.procs[jobid],
                    jobid,
                    "\n".join(format_exception(*sys.exc_info())),
                )
                self._clean_queue(jobid, graph)
                self.proc_pending[jobid] = False
            return True
        return False

    def _task_finished_cb(self, jobid, cached=False):
        """Extract outputs and assign to inputs of dependent tasks

        This is called when a job is completed.
        """
        logger.info(
            "[Job %d] %s (%s).",
            jobid,
            "Cached" if cached else "Completed",
            self.procs[jobid],
        )
        if self._status_callback:
            self._status_callback(self.procs[jobid], "end")
        # Update job and worker queues
        self.proc_pending[jobid] = False
        # update the job dependency structure
        rowview = self.depidx.getrowview(jobid)
        rowview[rowview.nonzero()] = 0
        if jobid not in self.mapnodesubids:
            self.refidx[self.refidx[:, jobid].nonzero()[0], jobid] = 0

    def _generate_dependency_list(self, graph):
        """Generates a dependency list for a list of graphs."""
        import networkx as nx

        self.procs, _ = topological_sort(graph)
        self.depidx = nx.to_scipy_sparse_matrix(
            graph, nodelist=self.procs, format="lil"
        )
        self.refidx = self.depidx.astype(int)
        self.proc_done = np.zeros(len(self.procs), dtype=bool)
        self.proc_pending = np.zeros(len(self.procs), dtype=bool)

    def _remove_node_deps(self, jobid, crashfile, graph):
        import networkx as nx

        try:
            dfs_preorder = nx.dfs_preorder
        except AttributeError:
            dfs_preorder = nx.dfs_preorder_nodes
        subnodes = [s for s in dfs_preorder(graph, self.procs[jobid])]
        for node in subnodes:
            idx = self.procs.index(node)
            self.proc_done[idx] = True
            self.proc_pending[idx] = False
        return dict(node=self.procs[jobid], dependents=subnodes, crashfile=crashfile)

    def _remove_node_dirs(self):
        """Removes directories whose outputs have already been used up"""
        if str2bool(self._config["execution"]["remove_node_directories"]):
            indices = np.nonzero((self.refidx.sum(axis=1) == 0).__array__())[0]
            for idx in indices:
                if idx in self.mapnodesubids:
                    continue
                if self.proc_done[idx] and (not self.proc_pending[idx]):
                    self.refidx[idx, idx] = -1
                    outdir = self.procs[idx].output_dir()
                    logger.info(
                        (
                            "[node dependencies finished] "
                            "removing node: %s from directory %s"
                        )
                        % (self.procs[idx]._id, outdir)
                    )
                    shutil.rmtree(outdir)


class SGELikeBatchManagerBase(DistributedPluginBase):
    """Execute workflow with SGE/OGE/PBS like batch system"""

    def __init__(self, template, plugin_args=None):
        super(SGELikeBatchManagerBase, self).__init__(plugin_args=plugin_args)
        self._template = template
        self._qsub_args = None
        if plugin_args:
            if "template" in plugin_args:
                self._template = plugin_args["template"]
                if os.path.isfile(self._template):
                    with open(self._template) as tpl_file:
                        self._template = tpl_file.read()
            if "qsub_args" in plugin_args:
                self._qsub_args = plugin_args["qsub_args"]
        self._pending = {}

    def _is_pending(self, taskid):
        """Check if a task is pending in the batch system"""
        raise NotImplementedError

    def _submit_batchtask(self, scriptfile, node):
        """Submit a task to the batch system"""
        raise NotImplementedError

    def _get_result(self, taskid):
        if taskid not in self._pending:
            raise Exception("Task %d not found" % taskid)
        if self._is_pending(taskid):
            return None
        node_dir = self._pending[taskid]
        # MIT HACK
        # on the pbs system at mit the parent node directory needs to be
        # accessed before internal directories become available. there
        # is a disconnect when the queueing engine knows a job is
        # finished to when the directories become statable.
        t = time()
        timeout = float(self._config["execution"]["job_finished_timeout"])
        timed_out = True
        while (time() - t) < timeout:
            try:
                glob(os.path.join(node_dir, "result_*.pklz")).pop()
                timed_out = False
                break
            except Exception as e:
                logger.debug(e)
            sleep(2)
        if timed_out:
            result_data = {"hostname": "unknown", "result": None, "traceback": None}
            results_file = None
            try:
                error_message = (
                    "Job id ({0}) finished or terminated, but "
                    "results file does not exist after ({1}) "
                    "seconds. Batch dir contains crashdump file "
                    "if node raised an exception.\n"
                    "Node working directory: ({2}) ".format(taskid, timeout, node_dir)
                )
                raise IOError(error_message)
            except IOError as e:
                result_data["traceback"] = "\n".join(format_exception(*sys.exc_info()))
        else:
            results_file = glob(os.path.join(node_dir, "result_*.pklz"))[0]
            result_data = load_resultfile(results_file)
        result_out = dict(result=None, traceback=None)
        if isinstance(result_data, dict):
            result_out["result"] = result_data["result"]
            result_out["traceback"] = result_data["traceback"]
            result_out["hostname"] = result_data["hostname"]
            if results_file:
                crash_file = os.path.join(node_dir, "crashstore.pklz")
                os.rename(results_file, crash_file)
        else:
            result_out["result"] = result_data
        return result_out

    def _submit_job(self, node, updatehash=False):
        """submit job and return taskid"""
        pyscript = create_pyscript(node, updatehash=updatehash)
        batch_dir, name = os.path.split(pyscript)
        name = ".".join(name.split(".")[:-1])
        batchscript = "\n".join(
            (self._template.rstrip("\n"), "%s %s" % (sys.executable, pyscript))
        )
        batchscriptfile = os.path.join(batch_dir, "batchscript_%s.sh" % name)
        with open(batchscriptfile, "wt") as fp:
            fp.writelines(batchscript)
        return self._submit_batchtask(batchscriptfile, node)

    def _clear_task(self, taskid):
        del self._pending[taskid]


class GraphPluginBase(PluginBase):
    """Base class for plugins that distribute graphs to workflows"""

    def __init__(self, plugin_args=None):
        if plugin_args and plugin_args.get("status_callback"):
            logger.warning(
                "status_callback not supported for Graph submission" " plugins"
            )
        super(GraphPluginBase, self).__init__(plugin_args=plugin_args)

    def run(self, graph, config, updatehash=False):
        import networkx as nx

        pyfiles = []
        dependencies = {}
        self._config = config
        nodes = list(nx.topological_sort(graph))
        logger.debug("Creating executable python files for each node")
        for idx, node in enumerate(nodes):
            pyfiles.append(
                create_pyscript(node, updatehash=updatehash, store_exception=False)
            )
            dependencies[idx] = [
                nodes.index(prevnode) for prevnode in list(graph.predecessors(node))
            ]
        self._submit_graph(pyfiles, dependencies, nodes)

    def _get_args(self, node, keywords):
        values = ()
        for keyword in keywords:
            value = getattr(self, "_" + keyword)
            if keyword == "template" and os.path.isfile(value):
                with open(value) as f:
                    value = f.read()
            if (
                hasattr(node, "plugin_args")
                and isinstance(node.plugin_args, dict)
                and keyword in node.plugin_args
            ):
                if keyword == "template" and os.path.isfile(node.plugin_args[keyword]):
                    with open(node.plugin_args[keyword]) as f:
                        tmp_value = f.read()
                else:
                    tmp_value = node.plugin_args[keyword]

                if "overwrite" in node.plugin_args and node.plugin_args["overwrite"]:
                    value = tmp_value
                else:
                    value += tmp_value
            values += (value,)
        return values

    def _submit_graph(self, pyfiles, dependencies, nodes):
        """
        pyfiles: list of files corresponding to a topological sort
        dependencies: dictionary of dependencies based on the toplogical sort
        """
        raise NotImplementedError

    def _get_result(self, taskid):
        if taskid not in self._pending:
            raise Exception("Task %d not found" % taskid)
        if self._is_pending(taskid):
            return None
        node_dir = self._pending[taskid]

        glob(os.path.join(node_dir, "result_*.pklz")).pop()

        results_file = glob(os.path.join(node_dir, "result_*.pklz"))[0]
        result_data = load_resultfile(results_file)
        result_out = dict(result=None, traceback=None)

        if isinstance(result_data, dict):
            result_out["result"] = result_data["result"]
            result_out["traceback"] = result_data["traceback"]
            result_out["hostname"] = result_data["hostname"]
            if results_file:
                crash_file = os.path.join(node_dir, "crashstore.pklz")
                os.rename(results_file, crash_file)
        else:
            result_out["result"] = result_data

        return result_out



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..gtract import gtractCoregBvalues


def test_gtractCoregBvalues_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        debugLevel=dict(
            argstr="--debugLevel %d",
        ),
        eddyCurrentCorrection=dict(
            argstr="--eddyCurrentCorrection ",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fixedVolume=dict(
            argstr="--fixedVolume %s",
            extensions=None,
        ),
        fixedVolumeIndex=dict(
            argstr="--fixedVolumeIndex %d",
        ),
        maximumStepSize=dict(
            argstr="--maximumStepSize %f",
        ),
        minimumStepSize=dict(
            argstr="--minimumStepSize %f",
        ),
        movingVolume=dict(
            argstr="--movingVolume %s",
            extensions=None,
        ),
        numberOfIterations=dict(
            argstr="--numberOfIterations %d",
        ),
        numberOfSpatialSamples=dict(
            argstr="--numberOfSpatialSamples %d",
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        outputTransform=dict(
            argstr="--outputTransform %s",
            hash_files=False,
        ),
        outputVolume=dict(
            argstr="--outputVolume %s",
            hash_files=False,
        ),
        registerB0Only=dict(
            argstr="--registerB0Only ",
        ),
        relaxationFactor=dict(
            argstr="--relaxationFactor %f",
        ),
        samplingPercentage=dict(
            argstr="--samplingPercentage %f",
        ),
        spatialScale=dict(
            argstr="--spatialScale %f",
        ),
    )
    inputs = gtractCoregBvalues.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_gtractCoregBvalues_outputs():
    output_map = dict(
        outputTransform=dict(
            extensions=None,
        ),
        outputVolume=dict(
            extensions=None,
        ),
    )
    outputs = gtractCoregBvalues.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Debug plugin
"""

import networkx as nx
from .base import PluginBase, logger


class DebugPlugin(PluginBase):
    """Execute workflow in series"""

    def __init__(self, plugin_args=None):
        super(DebugPlugin, self).__init__(plugin_args=plugin_args)
        if (
            plugin_args
            and "callable" in plugin_args
            and hasattr(plugin_args["callable"], "__call__")
        ):
            self._callable = plugin_args["callable"]
        else:
            raise ValueError("plugin_args must contain a callable function")

    def run(self, graph, config, updatehash=False):
        """Executes a pre-defined pipeline in a serial order.

        Parameters
        ----------

        graph : networkx digraph
            defines order of execution
        """

        if not isinstance(graph, nx.DiGraph):
            raise ValueError("Input must be a networkx digraph object")
        logger.info("Executing debug plugin")
        for node in nx.topological_sort(graph):
            self._callable(node, graph)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np
import nibabel as nb
import pytest

from nibabel.orientations import axcodes2ornt, ornt_transform

from ..image import _as_reoriented_backport, _orientations
from ... import LooseVersion

nibabel24 = LooseVersion(nb.__version__) >= LooseVersion("2.4.0")


@pytest.mark.skipif(not nibabel24, reason="Old nibabel - can't directly compare")
def test_reorientation_backport():
    pixdims = ((1, 1, 1), (2, 2, 3))
    data = np.random.normal(size=(17, 18, 19, 2))

    for pixdim in pixdims:
        # Generate a randomly rotated affine
        angles = np.random.uniform(-np.pi, np.pi, 3) * [1, 0.5, 1]
        rot = nb.eulerangles.euler2mat(*angles)
        scale = np.diag(pixdim)
        translation = np.array((17, 18, 19)) / 2
        affine = nb.affines.from_matvec(rot.dot(scale), translation)

        # Create image
        img = nb.Nifti1Image(data, affine)
        dim_info = {"freq": 0, "phase": 1, "slice": 2}
        img.header.set_dim_info(**dim_info)

        # Find a random, non-identity transform
        targ_ornt = orig_ornt = nb.io_orientation(affine)
        while np.array_equal(targ_ornt, orig_ornt):
            new_code = np.random.choice(_orientations)
            targ_ornt = axcodes2ornt(new_code)

        identity = ornt_transform(orig_ornt, orig_ornt)
        transform = ornt_transform(orig_ornt, targ_ornt)

        # Identity transform returns exact image
        assert img.as_reoriented(identity) is img
        assert _as_reoriented_backport(img, identity) is img

        reoriented_a = img.as_reoriented(transform)
        reoriented_b = _as_reoriented_backport(img, transform)

        flips_only = img.shape == reoriented_a.shape

        # Reorientation changes affine and data array
        assert not np.allclose(img.affine, reoriented_a.affine)
        assert not (
            flips_only and np.allclose(img.get_fdata(), reoriented_a.get_fdata())
        )
        # Dimension info changes iff axes are reordered
        assert flips_only == np.array_equal(
            img.header.get_dim_info(), reoriented_a.header.get_dim_info()
        )

        # Both approaches produce equivalent images
        assert np.allclose(reoriented_a.affine, reoriented_b.affine)
        assert np.array_equal(reoriented_a.get_fdata(), reoriented_b.get_fdata())
        assert np.array_equal(
            reoriented_a.header.get_dim_info(), reoriented_b.header.get_dim_info()
        )



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import MRIsCalc


def test_MRIsCalc_inputs():
    input_map = dict(
        action=dict(
            argstr="%s",
            mandatory=True,
            position=-2,
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file1=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-3,
        ),
        in_file2=dict(
            argstr="%s",
            extensions=None,
            position=-1,
            xor=["in_float", "in_int"],
        ),
        in_float=dict(
            argstr="%f",
            position=-1,
            xor=["in_file2", "in_int"],
        ),
        in_int=dict(
            argstr="%d",
            position=-1,
            xor=["in_file2", "in_float"],
        ),
        out_file=dict(
            argstr="-o %s",
            extensions=None,
            mandatory=True,
        ),
        subjects_dir=dict(),
    )
    inputs = MRIsCalc.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MRIsCalc_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = MRIsCalc.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utilities import HistogramMatchingFilter


def test_HistogramMatchingFilter_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        histogramAlgorithm=dict(
            argstr="--histogramAlgorithm %s",
        ),
        inputBinaryVolume=dict(
            argstr="--inputBinaryVolume %s",
            extensions=None,
        ),
        inputVolume=dict(
            argstr="--inputVolume %s",
            extensions=None,
        ),
        numberOfHistogramBins=dict(
            argstr="--numberOfHistogramBins %d",
        ),
        numberOfMatchPoints=dict(
            argstr="--numberOfMatchPoints %d",
        ),
        outputVolume=dict(
            argstr="--outputVolume %s",
            hash_files=False,
        ),
        referenceBinaryVolume=dict(
            argstr="--referenceBinaryVolume %s",
            extensions=None,
        ),
        referenceVolume=dict(
            argstr="--referenceVolume %s",
            extensions=None,
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
        writeHistogram=dict(
            argstr="--writeHistogram %s",
        ),
    )
    inputs = HistogramMatchingFilter.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_HistogramMatchingFilter_outputs():
    output_map = dict(
        outputVolume=dict(
            extensions=None,
        ),
    )
    outputs = HistogramMatchingFilter.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""DIPY is a computational neuroimaging tool for diffusion MRI."""
from .tracks import StreamlineTractography, TrackDensityMap
from .tensors import TensorMode, DTI
from .preprocess import Resample, Denoise
from .reconstruction import RESTORE, EstimateResponseSH, CSD
from .simulate import SimulateMultiTensor
from .anisotropic_power import APMQball



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import UnpackSDICOMDir


def test_UnpackSDICOMDir_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        config=dict(
            argstr="-cfg %s",
            extensions=None,
            mandatory=True,
            xor=("run_info", "config", "seq_config"),
        ),
        dir_structure=dict(
            argstr="-%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        log_file=dict(
            argstr="-log %s",
            extensions=None,
        ),
        no_info_dump=dict(
            argstr="-noinfodump",
        ),
        no_unpack_err=dict(
            argstr="-no-unpackerr",
        ),
        output_dir=dict(
            argstr="-targ %s",
        ),
        run_info=dict(
            argstr="-run %d %s %s %s",
            mandatory=True,
            xor=("run_info", "config", "seq_config"),
        ),
        scan_only=dict(
            argstr="-scanonly %s",
            extensions=None,
        ),
        seq_config=dict(
            argstr="-seqcfg %s",
            extensions=None,
            mandatory=True,
            xor=("run_info", "config", "seq_config"),
        ),
        source_dir=dict(
            argstr="-src %s",
            mandatory=True,
        ),
        spm_zeropad=dict(
            argstr="-nspmzeropad %d",
        ),
        subjects_dir=dict(),
    )
    inputs = UnpackSDICOMDir.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
import simplejson as json
import logging

import pytest
from unittest import mock

from .... import config
from ....testing import example_data
from ... import base as nib
from ..support import _inputs_help


def check_dict(ref_dict, tst_dict):
    """Compare dictionaries of inputs and and those loaded from json files"""

    def to_list(x):
        if isinstance(x, tuple):
            x = list(x)

        if isinstance(x, list):
            for i, xel in enumerate(x):
                x[i] = to_list(xel)

        return x

    failed_dict = {}
    for key, value in list(ref_dict.items()):
        newval = to_list(tst_dict[key])
        if newval != value:
            failed_dict[key] = (value, newval)
    return failed_dict


def test_Interface():
    assert nib.Interface.input_spec is None
    assert nib.Interface.output_spec is None
    with pytest.raises(NotImplementedError):
        nib.Interface()

    class DerivedInterface(nib.Interface):
        def __init__(self):
            pass

    nif = DerivedInterface()
    with pytest.raises(NotImplementedError):
        nif.run()
    with pytest.raises(NotImplementedError):
        nif.aggregate_outputs()
    with pytest.raises(NotImplementedError):
        nif._list_outputs()


def test_BaseInterface():
    config.set("monitoring", "enable", "0")

    assert nib.BaseInterface.help() is None

    class InputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int")
        goo = nib.traits.Int(desc="a random int", mandatory=True)
        moo = nib.traits.Int(desc="a random int", mandatory=False)
        hoo = nib.traits.Int(desc="a random int", usedefault=True)
        zoo = nib.File(desc="a file", copyfile=False)
        woo = nib.File(desc="a file", copyfile=True)

    class OutputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int")

    class DerivedInterface(nib.BaseInterface):
        input_spec = InputSpec
        resource_monitor = False

    assert DerivedInterface.help() is None
    assert "moo" in "".join(_inputs_help(DerivedInterface))
    assert DerivedInterface()._outputs() is None
    assert DerivedInterface().inputs.foo == nib.Undefined
    with pytest.raises(ValueError):
        DerivedInterface()._check_mandatory_inputs()
    assert DerivedInterface(goo=1)._check_mandatory_inputs() is None
    with pytest.raises(ValueError):
        DerivedInterface().run()
    with pytest.raises(NotImplementedError):
        DerivedInterface(goo=1).run()

    class DerivedInterface2(DerivedInterface):
        output_spec = OutputSpec

        def _run_interface(self, runtime):
            return runtime

    assert DerivedInterface2.help() is None
    assert DerivedInterface2()._outputs().foo == nib.Undefined
    with pytest.raises(NotImplementedError):
        DerivedInterface2(goo=1).run()

    default_inpu_spec = nib.BaseInterface.input_spec
    nib.BaseInterface.input_spec = None
    with pytest.raises(Exception):
        nib.BaseInterface()
    nib.BaseInterface.input_spec = default_inpu_spec


def test_BaseInterface_load_save_inputs(tmpdir):
    tmp_json = tmpdir.join("settings.json").strpath

    class InputSpec(nib.TraitedSpec):
        input1 = nib.traits.Int()
        input2 = nib.traits.Float()
        input3 = nib.traits.Bool()
        input4 = nib.traits.Str()

    class DerivedInterface(nib.BaseInterface):
        input_spec = InputSpec

        def __init__(self, **inputs):
            super(DerivedInterface, self).__init__(**inputs)

    inputs_dict = {"input1": 12, "input3": True, "input4": "some string"}
    bif = DerivedInterface(**inputs_dict)
    bif.save_inputs_to_json(tmp_json)
    bif2 = DerivedInterface()
    bif2.load_inputs_from_json(tmp_json)
    assert bif2.inputs.get_traitsfree() == inputs_dict

    bif3 = DerivedInterface(from_file=tmp_json)
    assert bif3.inputs.get_traitsfree() == inputs_dict

    inputs_dict2 = inputs_dict.copy()
    inputs_dict2.update({"input4": "some other string"})
    bif4 = DerivedInterface(from_file=tmp_json, input4=inputs_dict2["input4"])
    assert bif4.inputs.get_traitsfree() == inputs_dict2

    bif5 = DerivedInterface(input4=inputs_dict2["input4"])
    bif5.load_inputs_from_json(tmp_json, overwrite=False)
    assert bif5.inputs.get_traitsfree() == inputs_dict2

    bif6 = DerivedInterface(input4=inputs_dict2["input4"])
    bif6.load_inputs_from_json(tmp_json)
    assert bif6.inputs.get_traitsfree() == inputs_dict

    # test get hashval in a complex interface
    from nipype.interfaces.ants import Registration

    settings = example_data(example_data("smri_ants_registration_settings.json"))
    with open(settings) as setf:
        data_dict = json.load(setf)

    tsthash = Registration()
    tsthash.load_inputs_from_json(settings)
    assert {} == check_dict(data_dict, tsthash.inputs.get_traitsfree())

    tsthash2 = Registration(from_file=settings)
    assert {} == check_dict(data_dict, tsthash2.inputs.get_traitsfree())

    _, hashvalue = tsthash.inputs.get_hashval(hash_method="timestamp")
    assert "8562a5623562a871115eb14822ee8d02" == hashvalue


class MinVerInputSpec(nib.TraitedSpec):
    foo = nib.traits.Int(desc="a random int", min_ver="0.9")


class MaxVerInputSpec(nib.TraitedSpec):
    foo = nib.traits.Int(desc="a random int", max_ver="0.7")


def test_input_version_1():
    class DerivedInterface1(nib.BaseInterface):
        input_spec = MinVerInputSpec

    obj = DerivedInterface1()
    obj._check_version_requirements(obj.inputs)

    config.set("execution", "stop_on_unknown_version", True)

    with pytest.raises(ValueError) as excinfo:
        obj._check_version_requirements(obj.inputs)
    assert "no version information" in str(excinfo.value)

    config.set_default_config()


def test_input_version_2():
    class DerivedInterface1(nib.BaseInterface):
        input_spec = MinVerInputSpec
        _version = "0.8"

    obj = DerivedInterface1()
    obj.inputs.foo = 1
    with pytest.raises(Exception) as excinfo:
        obj._check_version_requirements(obj.inputs)
    assert "version 0.8 < required 0.9" in str(excinfo.value)


def test_input_version_3():
    class DerivedInterface1(nib.BaseInterface):
        input_spec = MinVerInputSpec
        _version = "0.10"

    obj = DerivedInterface1()
    obj._check_version_requirements(obj.inputs)


def test_input_version_4():
    class DerivedInterface1(nib.BaseInterface):
        input_spec = MinVerInputSpec
        _version = "0.9"

    obj = DerivedInterface1()
    obj.inputs.foo = 1
    obj._check_version_requirements(obj.inputs)


def test_input_version_5():
    class DerivedInterface2(nib.BaseInterface):
        input_spec = MaxVerInputSpec
        _version = "0.8"

    obj = DerivedInterface2()
    obj.inputs.foo = 1
    with pytest.raises(Exception) as excinfo:
        obj._check_version_requirements(obj.inputs)
    assert "version 0.8 > required 0.7" in str(excinfo.value)


def test_input_version_6():
    class DerivedInterface1(nib.BaseInterface):
        input_spec = MaxVerInputSpec
        _version = "0.7"

    obj = DerivedInterface1()
    obj.inputs.foo = 1
    obj._check_version_requirements(obj.inputs)


def test_input_version_missing(caplog):
    class DerivedInterface(nib.BaseInterface):
        class input_spec(nib.TraitedSpec):
            foo = nib.traits.Int(min_ver="0.9")
            bar = nib.traits.Int(max_ver="0.9")

        _version = "misparsed-garbage"

    obj = DerivedInterface()
    obj.inputs.foo = 1
    obj.inputs.bar = 1
    with caplog.at_level(logging.WARNING, logger="nipype.interface"):
        obj._check_version_requirements(obj.inputs)
    assert len(caplog.records) == 2


def test_input_version_missing_error(caplog):
    from nipype import config

    class DerivedInterface(nib.BaseInterface):
        class input_spec(nib.TraitedSpec):
            foo = nib.traits.Int(min_ver="0.9")
            bar = nib.traits.Int(max_ver="0.9")

        _version = "misparsed-garbage"

    obj1 = DerivedInterface(foo=1)
    obj2 = DerivedInterface(bar=1)
    with caplog.at_level(logging.WARNING, logger="nipype.interface"):
        with mock.patch.object(config, "getboolean", return_value=True):
            with pytest.raises(ValueError):
                obj1._check_version_requirements(obj1.inputs)
            with pytest.raises(ValueError):
                obj2._check_version_requirements(obj2.inputs)
    assert len(caplog.records) == 2


def test_unavailable_input():
    class WithInput(nib.BaseInterface):
        class input_spec(nib.TraitedSpec):
            foo = nib.traits.Int(3, usedefault=True, max_ver="0.5")

        _version = "0.4"

        def _run_interface(self, runtime):
            return runtime

    class WithoutInput(WithInput):
        _version = "0.6"

    has = WithInput()
    hasnt = WithoutInput()
    trying_anyway = WithoutInput(foo=3)
    assert has.inputs.foo == 3
    assert not nib.isdefined(hasnt.inputs.foo)
    assert trying_anyway.inputs.foo == 3

    has.run()
    hasnt.run()
    with pytest.raises(Exception):
        trying_anyway.run()

    # Still settable
    has.inputs.foo = 4
    hasnt.inputs.foo = 4
    trying_anyway.inputs.foo = 4
    assert has.inputs.foo == 4
    assert hasnt.inputs.foo == 4
    assert trying_anyway.inputs.foo == 4

    has.run()
    with pytest.raises(Exception):
        hasnt.run()
    with pytest.raises(Exception):
        trying_anyway.run()


def test_output_version():
    class InputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int")

    class OutputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int", min_ver="0.9")

    class DerivedInterface1(nib.BaseInterface):
        input_spec = InputSpec
        output_spec = OutputSpec
        _version = "0.10"
        resource_monitor = False

    obj = DerivedInterface1()
    assert obj._check_version_requirements(obj._outputs()) == []

    class InputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int")

    class OutputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int", min_ver="0.11")

    class DerivedInterface1(nib.BaseInterface):
        input_spec = InputSpec
        output_spec = OutputSpec
        _version = "0.10"
        resource_monitor = False

    obj = DerivedInterface1()
    assert obj._check_version_requirements(obj._outputs()) == ["foo"]

    class InputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int")

    class OutputSpec(nib.TraitedSpec):
        foo = nib.traits.Int(desc="a random int", min_ver="0.11")

    class DerivedInterface1(nib.BaseInterface):
        input_spec = InputSpec
        output_spec = OutputSpec
        _version = "0.10"
        resource_monitor = False

        def _run_interface(self, runtime):
            return runtime

        def _list_outputs(self):
            return {"foo": 1}

    obj = DerivedInterface1()
    with pytest.raises(KeyError):
        obj.run()


def test_Commandline():
    with pytest.raises(Exception):
        nib.CommandLine()
    ci = nib.CommandLine(command="which")
    assert ci.cmd == "which"
    assert ci.inputs.args == nib.Undefined
    ci2 = nib.CommandLine(command="which", args="ls")
    assert ci2.cmdline == "which ls"
    ci3 = nib.CommandLine(command="echo")
    ci3.resource_monitor = False
    ci3.inputs.environ = {"MYENV": "foo"}
    res = ci3.run()
    assert res.runtime.environ["MYENV"] == "foo"
    assert res.outputs is None

    class CommandLineInputSpec1(nib.CommandLineInputSpec):
        foo = nib.Str(argstr="%s", desc="a str")
        goo = nib.traits.Bool(argstr="-g", desc="a bool", position=0)
        hoo = nib.traits.List(argstr="-l %s", desc="a list")
        moo = nib.traits.List(argstr="-i %d...", desc="a repeated list", position=-1)
        noo = nib.traits.Int(argstr="-x %d", desc="an int")
        roo = nib.traits.Str(desc="not on command line")
        soo = nib.traits.Bool(argstr="-soo")

    nib.CommandLine.input_spec = CommandLineInputSpec1
    ci4 = nib.CommandLine(command="cmd")
    ci4.inputs.foo = "foo"
    ci4.inputs.goo = True
    ci4.inputs.hoo = ["a", "b"]
    ci4.inputs.moo = [1, 2, 3]
    ci4.inputs.noo = 0
    ci4.inputs.roo = "hello"
    ci4.inputs.soo = False
    cmd = ci4._parse_inputs()
    assert cmd[0] == "-g"
    assert cmd[-1] == "-i 1 -i 2 -i 3"
    assert "hello" not in " ".join(cmd)
    assert "-soo" not in " ".join(cmd)
    ci4.inputs.soo = True
    cmd = ci4._parse_inputs()
    assert "-soo" in " ".join(cmd)

    class CommandLineInputSpec2(nib.CommandLineInputSpec):
        foo = nib.File(argstr="%s", desc="a str", genfile=True)

    nib.CommandLine.input_spec = CommandLineInputSpec2
    ci5 = nib.CommandLine(command="cmd")
    with pytest.raises(NotImplementedError):
        ci5._parse_inputs()

    class DerivedClass(nib.CommandLine):
        input_spec = CommandLineInputSpec2

        def _gen_filename(self, name):
            return "filename"

    ci6 = DerivedClass(command="cmd")
    assert ci6._parse_inputs()[0] == "filename"
    nib.CommandLine.input_spec = nib.CommandLineInputSpec


def test_Commandline_environ(monkeypatch, tmpdir):
    from nipype import config

    config.set_default_config()

    tmpdir.chdir()
    monkeypatch.setitem(os.environ, "DISPLAY", ":1")
    # Test environment
    ci3 = nib.CommandLine(command="echo")
    res = ci3.run()
    assert res.runtime.environ["DISPLAY"] == ":1"

    # Test display_variable option
    monkeypatch.delitem(os.environ, "DISPLAY", raising=False)
    config.set("execution", "display_variable", ":3")
    res = ci3.run()
    assert "DISPLAY" not in ci3.inputs.environ
    assert "DISPLAY" not in res.runtime.environ

    # If the interface has _redirect_x then yes, it should be set
    ci3._redirect_x = True
    res = ci3.run()
    assert res.runtime.environ["DISPLAY"] == ":3"

    # Test overwrite
    monkeypatch.setitem(os.environ, "DISPLAY", ":1")
    ci3.inputs.environ = {"DISPLAY": ":2"}
    res = ci3.run()
    assert res.runtime.environ["DISPLAY"] == ":2"


def test_CommandLine_output(tmpdir):
    # Create one file
    tmpdir.chdir()
    file = tmpdir.join("foo.txt")
    file.write("123456\n")
    name = os.path.basename(file.strpath)

    ci = nib.CommandLine(command="ls -l")
    ci.terminal_output = "allatonce"
    res = ci.run()
    assert res.runtime.merged == ""
    assert name in res.runtime.stdout

    # Check stdout is written
    ci = nib.CommandLine(command="ls -l")
    ci.terminal_output = "file_stdout"
    res = ci.run()
    assert os.path.isfile("stdout.nipype")
    assert name in res.runtime.stdout
    tmpdir.join("stdout.nipype").remove(ignore_errors=True)

    # Check stderr is written
    ci = nib.CommandLine(command="ls -l")
    ci.terminal_output = "file_stderr"
    res = ci.run()
    assert os.path.isfile("stderr.nipype")
    tmpdir.join("stderr.nipype").remove(ignore_errors=True)

    # Check outputs are thrown away
    ci = nib.CommandLine(command="ls -l")
    ci.terminal_output = "none"
    res = ci.run()
    assert (
        res.runtime.stdout == ""
        and res.runtime.stderr == ""
        and res.runtime.merged == ""
    )

    # Check that new interfaces are set to default 'stream'
    ci = nib.CommandLine(command="ls -l")
    res = ci.run()
    assert ci.terminal_output == "stream"
    assert name in res.runtime.stdout and res.runtime.stderr == ""

    # Check only one file is generated
    ci = nib.CommandLine(command="ls -l")
    ci.terminal_output = "file"
    res = ci.run()
    assert os.path.isfile("output.nipype")
    assert (
        name in res.runtime.merged
        and res.runtime.stdout == ""
        and res.runtime.stderr == ""
    )
    tmpdir.join("output.nipype").remove(ignore_errors=True)

    # Check split files are generated
    ci = nib.CommandLine(command="ls -l")
    ci.terminal_output = "file_split"
    res = ci.run()
    assert os.path.isfile("stdout.nipype")
    assert os.path.isfile("stderr.nipype")
    assert name in res.runtime.stdout


def test_global_CommandLine_output(tmpdir):
    """Ensures CommandLine.set_default_terminal_output works"""
    from nipype.interfaces.fsl import BET

    ci = nib.CommandLine(command="ls -l")
    assert ci.terminal_output == "stream"  # default case

    ci = BET()
    assert ci.terminal_output == "stream"  # default case

    with mock.patch.object(nib.CommandLine, "_terminal_output"):
        nib.CommandLine.set_default_terminal_output("allatonce")
        ci = nib.CommandLine(command="ls -l")
        assert ci.terminal_output == "allatonce"

        nib.CommandLine.set_default_terminal_output("file")
        ci = nib.CommandLine(command="ls -l")
        assert ci.terminal_output == "file"

        # Check default affects derived interfaces
        ci = BET()
        assert ci.terminal_output == "file"


def test_CommandLine_prefix(tmpdir):
    tmpdir.chdir()
    oop = "out/of/path"
    os.makedirs(oop)

    script_name = "test_script.sh"
    script_path = os.path.join(oop, script_name)
    with open(script_path, "w") as script_f:
        script_f.write("#!/usr/bin/env bash\necho Success!")
    os.chmod(script_path, 0o755)

    ci = nib.CommandLine(command=script_name)
    with pytest.raises(IOError):
        ci.run()

    class OOPCLI(nib.CommandLine):
        _cmd_prefix = oop + "/"

    ci = OOPCLI(command=script_name)
    ci.run()

    class OOPShell(nib.CommandLine):
        _cmd_prefix = "bash {}/".format(oop)

    ci = OOPShell(command=script_name)
    ci.run()

    class OOPBadShell(nib.CommandLine):
        _cmd_prefix = "shell_dne {}/".format(oop)

    ci = OOPBadShell(command=script_name)
    with pytest.raises(IOError):
        ci.run()


def test_runtime_checks():
    class TestInterface(nib.BaseInterface):
        class input_spec(nib.TraitedSpec):
            a = nib.traits.Any()

        class output_spec(nib.TraitedSpec):
            b = nib.traits.Any()

        def _run_interface(self, runtime):
            return runtime

    class NoRuntime(TestInterface):
        def _run_interface(self, runtime):
            return None

    class BrokenRuntime(TestInterface):
        def _run_interface(self, runtime):
            del runtime.__dict__["cwd"]
            return runtime

    with pytest.raises(RuntimeError):
        NoRuntime().run()

    with pytest.raises(RuntimeError):
        BrokenRuntime().run()



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

import os

from ...base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)


class BRAINSConstellationModelerInputSpec(CommandLineInputSpec):
    verbose = traits.Bool(
        desc=",               Show more verbose output,             ",
        argstr="--verbose ",
    )
    inputTrainingList = File(
        desc=",               Setup file, giving all parameters for training up a template model for each landmark.,             ",
        exists=True,
        argstr="--inputTrainingList %s",
    )
    outputModel = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc=",               The full filename of the output model file.,             ",
        argstr="--outputModel %s",
    )
    saveOptimizedLandmarks = traits.Bool(
        desc=",               Flag to make a new subject-specific landmark definition file in the same format produced by Slicer3 with the optimized landmark (the detected RP, AC, and PC) in it.  Useful to tighten the variances in the ConstellationModeler.,             ",
        argstr="--saveOptimizedLandmarks ",
    )
    optimizedLandmarksFilenameExtender = traits.Str(
        desc=",                If the trainingList is (indexFullPathName) and contains landmark data filenames [path]/[filename].fcsv ,  make the optimized landmarks filenames out of [path]/[filename](thisExtender) and the optimized version of the input trainingList out of (indexFullPathName)(thisExtender) , when you rewrite all the landmarks according to the saveOptimizedLandmarks flag.,             ",
        argstr="--optimizedLandmarksFilenameExtender %s",
    )
    resultsDir = traits.Either(
        traits.Bool,
        Directory(),
        hash_files=False,
        desc=",               The directory for the results to be written.,             ",
        argstr="--resultsDir %s",
    )
    mspQualityLevel = traits.Int(
        desc=",                 Flag cotrols how agressive the MSP is estimated.  0=quick estimate (9 seconds), 1=normal estimate (11 seconds), 2=great estimate (22 seconds), 3=best estimate (58 seconds).,             ",
        argstr="--mspQualityLevel %d",
    )
    rescaleIntensities = traits.Bool(
        desc=",                 Flag to turn on rescaling image intensities on input.,             ",
        argstr="--rescaleIntensities ",
    )
    trimRescaledIntensities = traits.Float(
        desc=",                 Turn on clipping the rescaled image one-tailed on input.  Units of standard deviations above the mean.  Very large values are very permissive.  Non-positive value turns clipping off.  Defaults to removing 0.00001 of a normal tail above the mean.,             ",
        argstr="--trimRescaledIntensities %f",
    )
    rescaleIntensitiesOutputRange = InputMultiPath(
        traits.Int,
        desc=",                 This pair of integers gives the lower and upper bounds on the signal portion of the output image.  Out-of-field voxels are taken from BackgroundFillValue.,             ",
        sep=",",
        argstr="--rescaleIntensitiesOutputRange %s",
    )
    BackgroundFillValue = traits.Str(
        desc="Fill the background of image with specified short int value. Enter number or use BIGNEG for a large negative number.",
        argstr="--BackgroundFillValue %s",
    )
    writedebuggingImagesLevel = traits.Int(
        desc=",                 This flag controls if debugging images are produced.  By default value of 0 is no images.  Anything greater than zero will be increasing level of debugging images.,             ",
        argstr="--writedebuggingImagesLevel %d",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSConstellationModelerOutputSpec(TraitedSpec):
    outputModel = File(
        desc=",               The full filename of the output model file.,             ",
        exists=True,
    )
    resultsDir = Directory(
        desc=",               The directory for the results to be written.,             ",
        exists=True,
    )


class BRAINSConstellationModeler(SEMLikeCommandLine):
    """title: Generate Landmarks Model (BRAINS)

    category: Utilities.BRAINS

    description: Train up a model for BRAINSConstellationDetector
    """

    input_spec = BRAINSConstellationModelerInputSpec
    output_spec = BRAINSConstellationModelerOutputSpec
    _cmd = " BRAINSConstellationModeler "
    _outputs_filenames = {"outputModel": "outputModel.mdl", "resultsDir": "resultsDir"}
    _redirect_x = False


class landmarksConstellationWeightsInputSpec(CommandLineInputSpec):
    inputTrainingList = File(
        desc=",                 Setup file, giving all parameters for training up a Weight list for landmark.,             ",
        exists=True,
        argstr="--inputTrainingList %s",
    )
    inputTemplateModel = File(
        desc="User-specified template model.,             ",
        exists=True,
        argstr="--inputTemplateModel %s",
    )
    LLSModel = File(
        desc="Linear least squares model filename in HD5 format",
        exists=True,
        argstr="--LLSModel %s",
    )
    outputWeightsList = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc=",                 The filename of a csv file which is a list of landmarks and their corresponding weights.,             ",
        argstr="--outputWeightsList %s",
    )


class landmarksConstellationWeightsOutputSpec(TraitedSpec):
    outputWeightsList = File(
        desc=",                 The filename of a csv file which is a list of landmarks and their corresponding weights.,             ",
        exists=True,
    )


class landmarksConstellationWeights(SEMLikeCommandLine):
    """title: Generate Landmarks Weights (BRAINS)

    category: Utilities.BRAINS

    description: Train up a list of Weights for the Landmarks in BRAINSConstellationDetector
    """

    input_spec = landmarksConstellationWeightsInputSpec
    output_spec = landmarksConstellationWeightsOutputSpec
    _cmd = " landmarksConstellationWeights "
    _outputs_filenames = {"outputWeightsList": "outputWeightsList.wts"}
    _redirect_x = False


class BRAINSTrimForegroundInDirectionInputSpec(CommandLineInputSpec):
    inputVolume = File(
        desc="Input image to trim off the neck (and also air-filling noise.)",
        exists=True,
        argstr="--inputVolume %s",
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Output image with neck and air-filling noise trimmed isotropic image with AC at center of image.",
        argstr="--outputVolume %s",
    )
    directionCode = traits.Int(
        desc=",                 This flag chooses which dimension to compare.  The sign lets you flip direction.,             ",
        argstr="--directionCode %d",
    )
    otsuPercentileThreshold = traits.Float(
        desc=",                 This is a parameter to FindLargestForegroundFilledMask, which is employed to trim off air-filling noise.,             ",
        argstr="--otsuPercentileThreshold %f",
    )
    closingSize = traits.Int(
        desc=",                 This is a parameter to FindLargestForegroundFilledMask,             ",
        argstr="--closingSize %d",
    )
    headSizeLimit = traits.Float(
        desc=",                 Use this to vary from the command line our search for how much upper tissue is head for the center-of-mass calculation.  Units are CCs, not cubic millimeters.,             ",
        argstr="--headSizeLimit %f",
    )
    BackgroundFillValue = traits.Str(
        desc="Fill the background of image with specified short int value. Enter number or use BIGNEG for a large negative number.",
        argstr="--BackgroundFillValue %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSTrimForegroundInDirectionOutputSpec(TraitedSpec):
    outputVolume = File(
        desc="Output image with neck and air-filling noise trimmed isotropic image with AC at center of image.",
        exists=True,
    )


class BRAINSTrimForegroundInDirection(SEMLikeCommandLine):
    """title: Trim Foreground In Direction (BRAINS)

    category: Utilities.BRAINS

    description: This program will trim off the neck and also air-filling noise from the inputImage.

    version: 0.1

    documentation-url: http://www.nitrc.org/projects/art/
    """

    input_spec = BRAINSTrimForegroundInDirectionInputSpec
    output_spec = BRAINSTrimForegroundInDirectionOutputSpec
    _cmd = " BRAINSTrimForegroundInDirection "
    _outputs_filenames = {"outputVolume": "outputVolume.nii"}
    _redirect_x = False


class BRAINSLmkTransformInputSpec(CommandLineInputSpec):
    inputMovingLandmarks = File(
        desc="Input Moving Landmark list file in fcsv,             ",
        exists=True,
        argstr="--inputMovingLandmarks %s",
    )
    inputFixedLandmarks = File(
        desc="Input Fixed Landmark list file in fcsv,             ",
        exists=True,
        argstr="--inputFixedLandmarks %s",
    )
    outputAffineTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The filename for the estimated affine transform,             ",
        argstr="--outputAffineTransform %s",
    )
    inputMovingVolume = File(
        desc="The filename of input moving volume",
        exists=True,
        argstr="--inputMovingVolume %s",
    )
    inputReferenceVolume = File(
        desc="The filename of the reference volume",
        exists=True,
        argstr="--inputReferenceVolume %s",
    )
    outputResampledVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The filename of the output resampled volume",
        argstr="--outputResampledVolume %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSLmkTransformOutputSpec(TraitedSpec):
    outputAffineTransform = File(
        desc="The filename for the estimated affine transform,             ",
        exists=True,
    )
    outputResampledVolume = File(
        desc="The filename of the output resampled volume", exists=True
    )


class BRAINSLmkTransform(SEMLikeCommandLine):
    """title: Landmark Transform (BRAINS)

    category: Utilities.BRAINS

    description: This utility program estimates the affine transform to align the fixed landmarks to the moving landmarks, and then generate the resampled moving image to the same physical space as that of the reference image.

    version: 1.0

    documentation-url: http://www.nitrc.org/projects/brainscdetector/
    """

    input_spec = BRAINSLmkTransformInputSpec
    output_spec = BRAINSLmkTransformOutputSpec
    _cmd = " BRAINSLmkTransform "
    _outputs_filenames = {
        "outputResampledVolume": "outputResampledVolume.nii",
        "outputAffineTransform": "outputAffineTransform.h5",
    }
    _redirect_x = False


class BRAINSMushInputSpec(CommandLineInputSpec):
    inputFirstVolume = File(
        desc="Input image (1) for mixture optimization",
        exists=True,
        argstr="--inputFirstVolume %s",
    )
    inputSecondVolume = File(
        desc="Input image (2) for mixture optimization",
        exists=True,
        argstr="--inputSecondVolume %s",
    )
    inputMaskVolume = File(
        desc="Input label image for mixture optimization",
        exists=True,
        argstr="--inputMaskVolume %s",
    )
    outputWeightsFile = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Output Weights File",
        argstr="--outputWeightsFile %s",
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The MUSH image produced from the T1 and T2 weighted images",
        argstr="--outputVolume %s",
    )
    outputMask = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The brain volume mask generated from the MUSH image",
        argstr="--outputMask %s",
    )
    seed = InputMultiPath(
        traits.Int,
        desc="Seed Point for Brain Region Filling",
        sep=",",
        argstr="--seed %s",
    )
    desiredMean = traits.Float(
        desc="Desired mean within the mask for weighted sum of both images.",
        argstr="--desiredMean %f",
    )
    desiredVariance = traits.Float(
        desc="Desired variance within the mask for weighted sum of both images.",
        argstr="--desiredVariance %f",
    )
    lowerThresholdFactorPre = traits.Float(
        desc="Lower threshold factor for finding an initial brain mask",
        argstr="--lowerThresholdFactorPre %f",
    )
    upperThresholdFactorPre = traits.Float(
        desc="Upper threshold factor for finding an initial brain mask",
        argstr="--upperThresholdFactorPre %f",
    )
    lowerThresholdFactor = traits.Float(
        desc="Lower threshold factor for defining the brain mask",
        argstr="--lowerThresholdFactor %f",
    )
    upperThresholdFactor = traits.Float(
        desc="Upper threshold factor for defining the brain mask",
        argstr="--upperThresholdFactor %f",
    )
    boundingBoxSize = InputMultiPath(
        traits.Int,
        desc="Size of the cubic bounding box mask used when no brain mask is present",
        sep=",",
        argstr="--boundingBoxSize %s",
    )
    boundingBoxStart = InputMultiPath(
        traits.Int,
        desc="XYZ point-coordinate for the start of the cubic bounding box mask used when no brain mask is present",
        sep=",",
        argstr="--boundingBoxStart %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSMushOutputSpec(TraitedSpec):
    outputWeightsFile = File(desc="Output Weights File", exists=True)
    outputVolume = File(
        desc="The MUSH image produced from the T1 and T2 weighted images", exists=True
    )
    outputMask = File(
        desc="The brain volume mask generated from the MUSH image", exists=True
    )


class BRAINSMush(SEMLikeCommandLine):
    """title: Brain Extraction from T1/T2 image (BRAINS)

    category: Utilities.BRAINS

    description: This program: 1) generates a weighted mixture image optimizing the mean and variance and 2) produces a mask of the brain volume

    version: 0.1.0.$Revision: 1.4 $(alpha)

    documentation-url: http:://mri.radiology.uiowa.edu

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: This tool is a modification by Steven Dunn of a program developed by Greg Harris and Ron Pierson.

    acknowledgements: This work was developed by the University of Iowa Departments of Radiology and Psychiatry. This software was supported in part of NIH/NINDS award NS050568.
    """

    input_spec = BRAINSMushInputSpec
    output_spec = BRAINSMushOutputSpec
    _cmd = " BRAINSMush "
    _outputs_filenames = {
        "outputMask": "outputMask.nii.gz",
        "outputWeightsFile": "outputWeightsFile.txt",
        "outputVolume": "outputVolume.nii.gz",
    }
    _redirect_x = False


class BRAINSTransformConvertInputSpec(CommandLineInputSpec):
    inputTransform = File(exists=True, argstr="--inputTransform %s")
    referenceVolume = File(exists=True, argstr="--referenceVolume %s")
    outputTransformType = traits.Enum(
        "Affine",
        "VersorRigid",
        "ScaleVersor",
        "ScaleSkewVersor",
        "DisplacementField",
        "Same",
        desc="The target transformation type. Must be conversion-compatible with the input transform type",
        argstr="--outputTransformType %s",
    )
    outputPrecisionType = traits.Enum(
        "double",
        "float",
        desc="Precision type of the output transform. It can be either single precision or double precision",
        argstr="--outputPrecisionType %s",
    )
    displacementVolume = traits.Either(
        traits.Bool, File(), hash_files=False, argstr="--displacementVolume %s"
    )
    outputTransform = traits.Either(
        traits.Bool, File(), hash_files=False, argstr="--outputTransform %s"
    )


class BRAINSTransformConvertOutputSpec(TraitedSpec):
    displacementVolume = File(exists=True)
    outputTransform = File(exists=True)


class BRAINSTransformConvert(SEMLikeCommandLine):
    """title: BRAINS Transform Convert

    category: Utilities.BRAINS

    description: Convert ITK transforms to higher order transforms

    version: 1.0

    documentation-url: A utility to convert between transform file formats.

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: Hans J. Johnson,Kent Williams, Ali Ghayoor
    """

    input_spec = BRAINSTransformConvertInputSpec
    output_spec = BRAINSTransformConvertOutputSpec
    _cmd = " BRAINSTransformConvert "
    _outputs_filenames = {
        "displacementVolume": "displacementVolume.nii",
        "outputTransform": "outputTransform.mat",
    }
    _redirect_x = False


class landmarksConstellationAlignerInputSpec(CommandLineInputSpec):
    inputLandmarksPaired = File(
        desc="Input landmark file (.fcsv)",
        exists=True,
        argstr="--inputLandmarksPaired %s",
    )
    outputLandmarksPaired = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Output landmark file (.fcsv)",
        argstr="--outputLandmarksPaired %s",
    )


class landmarksConstellationAlignerOutputSpec(TraitedSpec):
    outputLandmarksPaired = File(desc="Output landmark file (.fcsv)", exists=True)


class landmarksConstellationAligner(SEMLikeCommandLine):
    """title: MidACPC Landmark Insertion

    category: Utilities.BRAINS

    description: This program converts the original landmark files to the acpc-aligned landmark files

    contributor: Ali Ghayoor
    """

    input_spec = landmarksConstellationAlignerInputSpec
    output_spec = landmarksConstellationAlignerOutputSpec
    _cmd = " landmarksConstellationAligner "
    _outputs_filenames = {"outputLandmarksPaired": "outputLandmarksPaired"}
    _redirect_x = False


class BRAINSEyeDetectorInputSpec(CommandLineInputSpec):
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )
    inputVolume = File(desc="The input volume", exists=True, argstr="--inputVolume %s")
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The output volume",
        argstr="--outputVolume %s",
    )
    debugDir = traits.Str(desc="A place for debug information", argstr="--debugDir %s")


class BRAINSEyeDetectorOutputSpec(TraitedSpec):
    outputVolume = File(desc="The output volume", exists=True)


class BRAINSEyeDetector(SEMLikeCommandLine):
    """title: Eye Detector (BRAINS)

    category: Utilities.BRAINS

    version: 1.0

    documentation-url: http://www.nitrc.org/projects/brainscdetector/
    """

    input_spec = BRAINSEyeDetectorInputSpec
    output_spec = BRAINSEyeDetectorOutputSpec
    _cmd = " BRAINSEyeDetector "
    _outputs_filenames = {"outputVolume": "outputVolume.nii"}
    _redirect_x = False


class BRAINSLinearModelerEPCAInputSpec(CommandLineInputSpec):
    inputTrainingList = File(
        desc="Input Training Landmark List Filename,             ",
        exists=True,
        argstr="--inputTrainingList %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSLinearModelerEPCAOutputSpec(TraitedSpec):
    pass


class BRAINSLinearModelerEPCA(SEMLikeCommandLine):
    """title: Landmark Linear Modeler (BRAINS)

    category: Utilities.BRAINS

    description: Training linear model using EPCA. Implementation based on my MS thesis, "A METHOD FOR AUTOMATED LANDMARK CONSTELLATION DETECTION USING EVOLUTIONARY PRINCIPAL COMPONENTS AND STATISTICAL SHAPE MODELS"

    version: 1.0

    documentation-url: http://www.nitrc.org/projects/brainscdetector/
    """

    input_spec = BRAINSLinearModelerEPCAInputSpec
    output_spec = BRAINSLinearModelerEPCAOutputSpec
    _cmd = " BRAINSLinearModelerEPCA "
    _outputs_filenames = {}
    _redirect_x = False


class BRAINSInitializedControlPointsInputSpec(CommandLineInputSpec):
    inputVolume = File(desc="Input Volume", exists=True, argstr="--inputVolume %s")
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Output Volume",
        argstr="--outputVolume %s",
    )
    splineGridSize = InputMultiPath(
        traits.Int,
        desc="The number of subdivisions of the BSpline Grid to be centered on the image space.  Each dimension must have at least 3 subdivisions for the BSpline to be correctly computed. ",
        sep=",",
        argstr="--splineGridSize %s",
    )
    permuteOrder = InputMultiPath(
        traits.Int,
        desc="The permutation order for the images.  The default is 0,1,2 (i.e. no permutation)",
        sep=",",
        argstr="--permuteOrder %s",
    )
    outputLandmarksFile = traits.Str(
        desc="Output filename", argstr="--outputLandmarksFile %s"
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSInitializedControlPointsOutputSpec(TraitedSpec):
    outputVolume = File(desc="Output Volume", exists=True)


class BRAINSInitializedControlPoints(SEMLikeCommandLine):
    """title: Initialized Control Points (BRAINS)

    category: Utilities.BRAINS

    description: Outputs bspline control points as landmarks

    version: 0.1.0.$Revision: 916 $(alpha)

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: Mark Scully

    acknowledgements: This work is part of the National Alliance for Medical Image Computing (NAMIC), funded by the National Institutes of Health through the NIH Roadmap for Medical Research, Grant U54 EB005149.  Additional support for Mark Scully and Hans Johnson at the University of Iowa.
    """

    input_spec = BRAINSInitializedControlPointsInputSpec
    output_spec = BRAINSInitializedControlPointsOutputSpec
    _cmd = " BRAINSInitializedControlPoints "
    _outputs_filenames = {"outputVolume": "outputVolume.nii"}
    _redirect_x = False


class CleanUpOverlapLabelsInputSpec(CommandLineInputSpec):
    inputBinaryVolumes = InputMultiPath(
        File(exists=True),
        desc="The list of binary images to be checked and cleaned up. Order is important. Binary volume given first always wins out. ",
        argstr="--inputBinaryVolumes %s...",
    )
    outputBinaryVolumes = traits.Either(
        traits.Bool,
        InputMultiPath(File()),
        hash_files=False,
        desc="The output label map images, with integer values in it. Each label value specified in the inputLabels is combined into this output label map volume",
        argstr="--outputBinaryVolumes %s...",
    )


class CleanUpOverlapLabelsOutputSpec(TraitedSpec):
    outputBinaryVolumes = OutputMultiPath(
        File(exists=True),
        desc="The output label map images, with integer values in it. Each label value specified in the inputLabels is combined into this output label map volume",
    )


class CleanUpOverlapLabels(SEMLikeCommandLine):
    """title: Clean Up Overla Labels

    category: Utilities.BRAINS

    description: Take a series of input binary images and clean up for those overlapped area. Binary volumes given first always wins out

    version: 0.1.0

    contributor: Eun Young Kim
    """

    input_spec = CleanUpOverlapLabelsInputSpec
    output_spec = CleanUpOverlapLabelsOutputSpec
    _cmd = " CleanUpOverlapLabels "
    _outputs_filenames = {"outputBinaryVolumes": "outputBinaryVolumes.nii"}
    _redirect_x = False


class BRAINSClipInferiorInputSpec(CommandLineInputSpec):
    inputVolume = File(
        desc="Input image to make a clipped short int copy from.",
        exists=True,
        argstr="--inputVolume %s",
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Output image, a short int copy of the upper portion of the input image, filled with BackgroundFillValue.",
        argstr="--outputVolume %s",
    )
    acLowerBound = traits.Float(
        desc=",                 When the input image to the output image, replace the image with the BackgroundFillValue everywhere below the plane This Far in physical units (millimeters) below (inferior to) the AC point (assumed to be the voxel field middle.)  The oversize default was chosen to have no effect.  Based on visualizing a thousand masks in the IPIG study, we recommend a limit no smaller than 80.0 mm.,             ",
        argstr="--acLowerBound %f",
    )
    BackgroundFillValue = traits.Str(
        desc="Fill the background of image with specified short int value. Enter number or use BIGNEG for a large negative number.",
        argstr="--BackgroundFillValue %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSClipInferiorOutputSpec(TraitedSpec):
    outputVolume = File(
        desc="Output image, a short int copy of the upper portion of the input image, filled with BackgroundFillValue.",
        exists=True,
    )


class BRAINSClipInferior(SEMLikeCommandLine):
    """title: Clip Inferior of Center of Brain (BRAINS)

    category: Utilities.BRAINS

    description: This program will read the inputVolume as a short int image, write the BackgroundFillValue everywhere inferior to the lower bound, and write the resulting clipped short int image in the outputVolume.

    version: 1.0
    """

    input_spec = BRAINSClipInferiorInputSpec
    output_spec = BRAINSClipInferiorOutputSpec
    _cmd = " BRAINSClipInferior "
    _outputs_filenames = {"outputVolume": "outputVolume.nii"}
    _redirect_x = False


class GenerateLabelMapFromProbabilityMapInputSpec(CommandLineInputSpec):
    inputVolumes = InputMultiPath(
        File(exists=True),
        desc="The Input probaiblity images to be computed for lable maps",
        argstr="--inputVolumes %s...",
    )
    outputLabelVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="The Input binary image for region of interest",
        argstr="--outputLabelVolume %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class GenerateLabelMapFromProbabilityMapOutputSpec(TraitedSpec):
    outputLabelVolume = File(
        desc="The Input binary image for region of interest", exists=True
    )


class GenerateLabelMapFromProbabilityMap(SEMLikeCommandLine):
    """title: Label Map from Probability Images

    category: Utilities.BRAINS

    description: Given a list of probability maps for labels, create a discrete label map where only the highest probability region is used for the labeling.

    version: 0.1

    contributor: University of Iowa Department of Psychiatry, http:://www.psychiatry.uiowa.edu
    """

    input_spec = GenerateLabelMapFromProbabilityMapInputSpec
    output_spec = GenerateLabelMapFromProbabilityMapOutputSpec
    _cmd = " GenerateLabelMapFromProbabilityMap "
    _outputs_filenames = {"outputLabelVolume": "outputLabelVolume.nii.gz"}
    _redirect_x = False


class BRAINSAlignMSPInputSpec(CommandLineInputSpec):
    inputVolume = File(
        desc=",         The Image to be resampled,       ",
        exists=True,
        argstr="--inputVolume %s",
    )
    OutputresampleMSP = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc=",         The image to be output.,       ",
        argstr="--OutputresampleMSP %s",
    )
    verbose = traits.Bool(
        desc=",         Show more verbose output,       ", argstr="--verbose "
    )
    resultsDir = traits.Either(
        traits.Bool,
        Directory(),
        hash_files=False,
        desc=",         The directory for the results to be written.,       ",
        argstr="--resultsDir %s",
    )
    writedebuggingImagesLevel = traits.Int(
        desc=",           This flag controls if debugging images are produced.  By default value of 0 is no images.  Anything greater than zero will be increasing level of debugging images.,       ",
        argstr="--writedebuggingImagesLevel %d",
    )
    mspQualityLevel = traits.Int(
        desc=",           Flag cotrols how agressive the MSP is estimated.  0=quick estimate (9 seconds), 1=normal estimate (11 seconds), 2=great estimate (22 seconds), 3=best estimate (58 seconds).,       ",
        argstr="--mspQualityLevel %d",
    )
    rescaleIntensities = traits.Bool(
        desc=",           Flag to turn on rescaling image intensities on input.,       ",
        argstr="--rescaleIntensities ",
    )
    trimRescaledIntensities = traits.Float(
        desc=",           Turn on clipping the rescaled image one-tailed on input.  Units of standard deviations above the mean.  Very large values are very permissive.  Non-positive value turns clipping off.  Defaults to removing 0.00001 of a normal tail above the mean.,       ",
        argstr="--trimRescaledIntensities %f",
    )
    rescaleIntensitiesOutputRange = InputMultiPath(
        traits.Int,
        desc=",           This pair of integers gives the lower and upper bounds on the signal portion of the output image.  Out-of-field voxels are taken from BackgroundFillValue.,       ",
        sep=",",
        argstr="--rescaleIntensitiesOutputRange %s",
    )
    BackgroundFillValue = traits.Str(
        desc="Fill the background of image with specified short int value. Enter number or use BIGNEG for a large negative number.",
        argstr="--BackgroundFillValue %s",
    )
    interpolationMode = traits.Enum(
        "NearestNeighbor",
        "Linear",
        "ResampleInPlace",
        "BSpline",
        "WindowedSinc",
        "Hamming",
        "Cosine",
        "Welch",
        "Lanczos",
        "Blackman",
        desc="Type of interpolation to be used when applying transform to moving volume.  Options are Linear, ResampleInPlace, NearestNeighbor, BSpline, or WindowedSinc",
        argstr="--interpolationMode %s",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class BRAINSAlignMSPOutputSpec(TraitedSpec):
    OutputresampleMSP = File(
        desc=",         The image to be output.,       ", exists=True
    )
    resultsDir = Directory(
        desc=",         The directory for the results to be written.,       ",
        exists=True,
    )


class BRAINSAlignMSP(SEMLikeCommandLine):
    """title: Align Mid Saggital Brain (BRAINS)

    category: Utilities.BRAINS

    description: Resample an image into ACPC alignement ACPCDetect
    """

    input_spec = BRAINSAlignMSPInputSpec
    output_spec = BRAINSAlignMSPOutputSpec
    _cmd = " BRAINSAlignMSP "
    _outputs_filenames = {
        "OutputresampleMSP": "OutputresampleMSP.nii",
        "resultsDir": "resultsDir",
    }
    _redirect_x = False


class BRAINSLandmarkInitializerInputSpec(CommandLineInputSpec):
    inputFixedLandmarkFilename = File(
        desc="input fixed landmark. *.fcsv",
        exists=True,
        argstr="--inputFixedLandmarkFilename %s",
    )
    inputMovingLandmarkFilename = File(
        desc="input moving landmark. *.fcsv",
        exists=True,
        argstr="--inputMovingLandmarkFilename %s",
    )
    inputWeightFilename = File(
        desc="Input weight file name for landmarks. Higher weighted landmark will be considered more heavily. Weights are propotional, that is the magnitude of weights will be normalized by its minimum and maximum value. ",
        exists=True,
        argstr="--inputWeightFilename %s",
    )
    outputTransformFilename = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="output transform file name (ex: ./outputTransform.mat) ",
        argstr="--outputTransformFilename %s",
    )


class BRAINSLandmarkInitializerOutputSpec(TraitedSpec):
    outputTransformFilename = File(
        desc="output transform file name (ex: ./outputTransform.mat) ", exists=True
    )


class BRAINSLandmarkInitializer(SEMLikeCommandLine):
    """title: BRAINSLandmarkInitializer

    category: Utilities.BRAINS

    description: Create transformation file (*mat) from a pair of landmarks (*fcsv) files.

    version: 1.0

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: Eunyoung Regina Kim
    """

    input_spec = BRAINSLandmarkInitializerInputSpec
    output_spec = BRAINSLandmarkInitializerOutputSpec
    _cmd = " BRAINSLandmarkInitializer "
    _outputs_filenames = {"outputTransformFilename": "outputTransformFilename"}
    _redirect_x = False


class insertMidACPCpointInputSpec(CommandLineInputSpec):
    inputLandmarkFile = File(
        desc="Input landmark file (.fcsv)", exists=True, argstr="--inputLandmarkFile %s"
    )
    outputLandmarkFile = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Output landmark file (.fcsv)",
        argstr="--outputLandmarkFile %s",
    )


class insertMidACPCpointOutputSpec(TraitedSpec):
    outputLandmarkFile = File(desc="Output landmark file (.fcsv)", exists=True)


class insertMidACPCpoint(SEMLikeCommandLine):
    """title: MidACPC Landmark Insertion

    category: Utilities.BRAINS

    description: This program gets a landmark fcsv file and adds a new landmark as the midpoint between AC and PC points to the output landmark fcsv file

    contributor: Ali Ghayoor
    """

    input_spec = insertMidACPCpointInputSpec
    output_spec = insertMidACPCpointOutputSpec
    _cmd = " insertMidACPCpoint "
    _outputs_filenames = {"outputLandmarkFile": "outputLandmarkFile"}
    _redirect_x = False


class BRAINSSnapShotWriterInputSpec(CommandLineInputSpec):
    inputVolumes = InputMultiPath(
        File(exists=True),
        desc="Input image volume list to be extracted as 2D image. Multiple input is possible. At least one input is required.",
        argstr="--inputVolumes %s...",
    )
    inputBinaryVolumes = InputMultiPath(
        File(exists=True),
        desc="Input mask (binary) volume list to be extracted as 2D image. Multiple input is possible.",
        argstr="--inputBinaryVolumes %s...",
    )
    inputSliceToExtractInPhysicalPoint = InputMultiPath(
        traits.Float,
        desc="2D slice number of input images. For autoWorkUp output, which AC-PC aligned, 0,0,0 will be the center.",
        sep=",",
        argstr="--inputSliceToExtractInPhysicalPoint %s",
    )
    inputSliceToExtractInIndex = InputMultiPath(
        traits.Int,
        desc="2D slice number of input images. For size of 256*256*256 image, 128 is usually used.",
        sep=",",
        argstr="--inputSliceToExtractInIndex %s",
    )
    inputSliceToExtractInPercent = InputMultiPath(
        traits.Int,
        desc="2D slice number of input images. Percentage input from 0%-100%. (ex. --inputSliceToExtractInPercent 50,50,50",
        sep=",",
        argstr="--inputSliceToExtractInPercent %s",
    )
    inputPlaneDirection = InputMultiPath(
        traits.Int,
        desc="Plane to display. In general, 0=saggital, 1=coronal, and 2=axial plane.",
        sep=",",
        argstr="--inputPlaneDirection %s",
    )
    outputFilename = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="2D file name of input images. Required.",
        argstr="--outputFilename %s",
    )


class BRAINSSnapShotWriterOutputSpec(TraitedSpec):
    outputFilename = File(desc="2D file name of input images. Required.", exists=True)


class BRAINSSnapShotWriter(SEMLikeCommandLine):
    """title: BRAINSSnapShotWriter

    category: Utilities.BRAINS

    description: Create 2D snapshot of input images. Mask images are color-coded

    version: 1.0

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: Eunyoung Regina Kim
    """

    input_spec = BRAINSSnapShotWriterInputSpec
    output_spec = BRAINSSnapShotWriterOutputSpec
    _cmd = " BRAINSSnapShotWriter "
    _outputs_filenames = {"outputFilename": "outputFilename"}
    _redirect_x = False


class JointHistogramInputSpec(CommandLineInputSpec):
    inputVolumeInXAxis = File(
        desc="The Input image to be computed for statistics",
        exists=True,
        argstr="--inputVolumeInXAxis %s",
    )
    inputVolumeInYAxis = File(
        desc="The Input image to be computed for statistics",
        exists=True,
        argstr="--inputVolumeInYAxis %s",
    )
    inputMaskVolumeInXAxis = File(
        desc="Input mask volume for inputVolumeInXAxis. Histogram will be computed just for the masked region",
        exists=True,
        argstr="--inputMaskVolumeInXAxis %s",
    )
    inputMaskVolumeInYAxis = File(
        desc="Input mask volume for inputVolumeInYAxis. Histogram will be computed just for the masked region",
        exists=True,
        argstr="--inputMaskVolumeInYAxis %s",
    )
    outputJointHistogramImage = traits.Str(
        desc=" output joint histogram image file name. Histogram is usually 2D image. ",
        argstr="--outputJointHistogramImage %s",
    )
    verbose = traits.Bool(
        desc=" print debugging information,       ", argstr="--verbose "
    )


class JointHistogramOutputSpec(TraitedSpec):
    pass


class JointHistogram(SEMLikeCommandLine):
    """title: Write Out Image Intensities

    category: Utilities.BRAINS

    description: For Analysis

    version: 0.1

    contributor: University of Iowa Department of Psychiatry, http:://www.psychiatry.uiowa.edu
    """

    input_spec = JointHistogramInputSpec
    output_spec = JointHistogramOutputSpec
    _cmd = " JointHistogram "
    _outputs_filenames = {}
    _redirect_x = False


class ShuffleVectorsModuleInputSpec(CommandLineInputSpec):
    inputVectorFileBaseName = File(
        desc="input vector file name prefix. Usually end with .txt and header file has prost fix of .txt.hdr",
        exists=True,
        argstr="--inputVectorFileBaseName %s",
    )
    outputVectorFileBaseName = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="output vector file name prefix. Usually end with .txt and header file has prost fix of .txt.hdr",
        argstr="--outputVectorFileBaseName %s",
    )
    resampleProportion = traits.Float(
        desc="downsample size of 1 will be the same size as the input images, downsample size of 3 will throw 2/3 the vectors away.",
        argstr="--resampleProportion %f",
    )


class ShuffleVectorsModuleOutputSpec(TraitedSpec):
    outputVectorFileBaseName = File(
        desc="output vector file name prefix. Usually end with .txt and header file has prost fix of .txt.hdr",
        exists=True,
    )


class ShuffleVectorsModule(SEMLikeCommandLine):
    """title: ShuffleVectors

    category: Utilities.BRAINS

    description: Automatic Segmentation using neural networks

    version: 1.0

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: Hans Johnson
    """

    input_spec = ShuffleVectorsModuleInputSpec
    output_spec = ShuffleVectorsModuleOutputSpec
    _cmd = " ShuffleVectorsModule "
    _outputs_filenames = {"outputVectorFileBaseName": "outputVectorFileBaseName"}
    _redirect_x = False


class ImageRegionPlotterInputSpec(CommandLineInputSpec):
    inputVolume1 = File(
        desc="The Input image to be computed for statistics",
        exists=True,
        argstr="--inputVolume1 %s",
    )
    inputVolume2 = File(
        desc="The Input image to be computed for statistics",
        exists=True,
        argstr="--inputVolume2 %s",
    )
    inputBinaryROIVolume = File(
        desc="The Input binary image for region of interest",
        exists=True,
        argstr="--inputBinaryROIVolume %s",
    )
    inputLabelVolume = File(
        desc="The Label Image", exists=True, argstr="--inputLabelVolume %s"
    )
    numberOfHistogramBins = traits.Int(
        desc=" the number of histogram levels", argstr="--numberOfHistogramBins %d"
    )
    outputJointHistogramData = traits.Str(
        desc=" output data file name", argstr="--outputJointHistogramData %s"
    )
    useROIAUTO = traits.Bool(
        desc=" Use ROIAUTO to compute region of interest. This cannot be used with inputLabelVolume",
        argstr="--useROIAUTO ",
    )
    useIntensityForHistogram = traits.Bool(
        desc=" Create Intensity Joint Histogram instead of Quantile Joint Histogram",
        argstr="--useIntensityForHistogram ",
    )
    verbose = traits.Bool(
        desc=" print debugging information,       ", argstr="--verbose "
    )


class ImageRegionPlotterOutputSpec(TraitedSpec):
    pass


class ImageRegionPlotter(SEMLikeCommandLine):
    """title: Write Out Image Intensities

    category: Utilities.BRAINS

    description: For Analysis

    version: 0.1

    contributor: University of Iowa Department of Psychiatry, http:://www.psychiatry.uiowa.edu
    """

    input_spec = ImageRegionPlotterInputSpec
    output_spec = ImageRegionPlotterOutputSpec
    _cmd = " ImageRegionPlotter "
    _outputs_filenames = {}
    _redirect_x = False


class fcsv_to_hdf5InputSpec(CommandLineInputSpec):
    versionID = traits.Str(
        desc=",         Current version ID. It should be match with the version of BCD that will be using the output model file,       ",
        argstr="--versionID %s",
    )
    landmarksInformationFile = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc=",         name of HDF5 file to write matrices into,       ",
        argstr="--landmarksInformationFile %s",
    )
    landmarkTypesList = File(
        desc=",         file containing list of landmark types,       ",
        exists=True,
        argstr="--landmarkTypesList %s",
    )
    modelFile = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc=",         name of HDF5 file containing BRAINSConstellationDetector Model file (LLSMatrices, LLSMeans and LLSSearchRadii),       ",
        argstr="--modelFile %s",
    )
    landmarkGlobPattern = traits.Str(
        desc="Glob pattern to select fcsv files", argstr="--landmarkGlobPattern %s"
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use.",
        argstr="--numberOfThreads %d",
    )


class fcsv_to_hdf5OutputSpec(TraitedSpec):
    landmarksInformationFile = File(
        desc=",         name of HDF5 file to write matrices into,       ", exists=True
    )
    modelFile = File(
        desc=",         name of HDF5 file containing BRAINSConstellationDetector Model file (LLSMatrices, LLSMeans and LLSSearchRadii),       ",
        exists=True,
    )


class fcsv_to_hdf5(SEMLikeCommandLine):
    """title: fcsv_to_hdf5 (BRAINS)

    category: Utilities.BRAINS

    description: Convert a collection of fcsv files to a HDF5 format file
    """

    input_spec = fcsv_to_hdf5InputSpec
    output_spec = fcsv_to_hdf5OutputSpec
    _cmd = " fcsv_to_hdf5 "
    _outputs_filenames = {
        "modelFile": "modelFile",
        "landmarksInformationFile": "landmarksInformationFile.h5",
    }
    _redirect_x = False


class FindCenterOfBrainInputSpec(CommandLineInputSpec):
    inputVolume = File(
        desc="The image in which to find the center.",
        exists=True,
        argstr="--inputVolume %s",
    )
    imageMask = File(exists=True, argstr="--imageMask %s")
    clippedImageMask = traits.Either(
        traits.Bool, File(), hash_files=False, argstr="--clippedImageMask %s"
    )
    maximize = traits.Bool(argstr="--maximize ")
    axis = traits.Int(argstr="--axis %d")
    otsuPercentileThreshold = traits.Float(argstr="--otsuPercentileThreshold %f")
    closingSize = traits.Int(argstr="--closingSize %d")
    headSizeLimit = traits.Float(argstr="--headSizeLimit %f")
    headSizeEstimate = traits.Float(argstr="--headSizeEstimate %f")
    backgroundValue = traits.Int(argstr="--backgroundValue %d")
    generateDebugImages = traits.Bool(argstr="--generateDebugImages ")
    debugDistanceImage = traits.Either(
        traits.Bool, File(), hash_files=False, argstr="--debugDistanceImage %s"
    )
    debugGridImage = traits.Either(
        traits.Bool, File(), hash_files=False, argstr="--debugGridImage %s"
    )
    debugAfterGridComputationsForegroundImage = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        argstr="--debugAfterGridComputationsForegroundImage %s",
    )
    debugClippedImageMask = traits.Either(
        traits.Bool, File(), hash_files=False, argstr="--debugClippedImageMask %s"
    )
    debugTrimmedImage = traits.Either(
        traits.Bool, File(), hash_files=False, argstr="--debugTrimmedImage %s"
    )


class FindCenterOfBrainOutputSpec(TraitedSpec):
    clippedImageMask = File(exists=True)
    debugDistanceImage = File(exists=True)
    debugGridImage = File(exists=True)
    debugAfterGridComputationsForegroundImage = File(exists=True)
    debugClippedImageMask = File(exists=True)
    debugTrimmedImage = File(exists=True)


class FindCenterOfBrain(SEMLikeCommandLine):
    """title: Center Of Brain (BRAINS)

    category: Utilities.BRAINS

    description: Finds the center point of a brain

    version: 3.0.0

    license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

    contributor: Hans J. Johnson, hans-johnson -at- uiowa.edu, http://wwww.psychiatry.uiowa.edu

    acknowledgements: Hans Johnson(1,3,4); Kent Williams(1);  (1=University of Iowa Department of Psychiatry, 3=University of Iowa Department of Biomedical Engineering, 4=University of Iowa Department of Electrical and Computer Engineering
    """

    input_spec = FindCenterOfBrainInputSpec
    output_spec = FindCenterOfBrainOutputSpec
    _cmd = " FindCenterOfBrain "
    _outputs_filenames = {
        "debugClippedImageMask": "debugClippedImageMask.nii",
        "debugTrimmedImage": "debugTrimmedImage.nii",
        "debugDistanceImage": "debugDistanceImage.nii",
        "debugGridImage": "debugGridImage.nii",
        "clippedImageMask": "clippedImageMask.nii",
        "debugAfterGridComputationsForegroundImage": "debugAfterGridComputationsForegroundImage.nii",
    }
    _redirect_x = False



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
import tempfile

import numpy as np

import pytest
from nipype.testing import example_data
import nipype.interfaces.nitime as nitime

no_nitime = not nitime.analysis.have_nitime
display_available = "DISPLAY" in os.environ and os.environ["DISPLAY"]


@pytest.mark.skipif(no_nitime, reason="nitime is not installed")
def test_read_csv():
    """Test that reading the data from csv file gives you back a reasonable
    time-series object"""
    CA = nitime.CoherenceAnalyzer()
    CA.inputs.TR = 1.89  # bogus value just to pass traits test
    CA.inputs.in_file = example_data("fmri_timeseries_nolabels.csv")
    with pytest.raises(ValueError):
        CA._read_csv()

    CA.inputs.in_file = example_data("fmri_timeseries.csv")
    data, roi_names = CA._read_csv()
    assert data[0][0] == 10125.9
    assert roi_names[0] == "WM"


@pytest.mark.skipif(no_nitime, reason="nitime is not installed")
def test_coherence_analysis(tmpdir):
    """Test that the coherence analyzer works"""
    import nitime.analysis as nta
    import nitime.timeseries as ts

    tmpdir.chdir()
    # This is the nipype interface analysis:
    CA = nitime.CoherenceAnalyzer()
    CA.inputs.TR = 1.89
    CA.inputs.in_file = example_data("fmri_timeseries.csv")
    if display_available:
        tmp_png = tempfile.mkstemp(suffix=".png")[1]
        CA.inputs.output_figure_file = tmp_png
    tmp_csv = tempfile.mkstemp(suffix=".csv")[1]
    CA.inputs.output_csv_file = tmp_csv

    o = CA.run()
    assert o.outputs.coherence_array.shape == (31, 31)

    # This is the nitime analysis:
    TR = 1.89
    data_rec = np.recfromcsv(example_data("fmri_timeseries.csv"))
    roi_names = np.array(data_rec.dtype.names)
    n_samples = data_rec.shape[0]
    data = np.zeros((len(roi_names), n_samples))

    for n_idx, roi in enumerate(roi_names):
        data[n_idx] = data_rec[roi]

    T = ts.TimeSeries(data, sampling_interval=TR)

    assert (CA._csv2ts().data == T.data).all()

    T.metadata["roi"] = roi_names
    C = nta.CoherenceAnalyzer(
        T,
        method=dict(
            this_method="welch", NFFT=CA.inputs.NFFT, n_overlap=CA.inputs.n_overlap
        ),
    )

    freq_idx = np.where(
        (C.frequencies > CA.inputs.frequency_range[0])
        * (C.frequencies < CA.inputs.frequency_range[1])
    )[0]

    # Extract the coherence and average across these frequency bands:
    # Averaging is done on the last dimension
    coh = np.mean(C.coherence[:, :, freq_idx], -1)

    assert (o.outputs.coherence_array == coh).all()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import pytest

from ....testing import example_data
from ...niftyreg import get_custom_path
from ...niftyreg.tests.test_regutils import no_nifty_tool
from .. import EM


@pytest.mark.skipif(no_nifty_tool(cmd="seg_EM"), reason="niftyseg is not installed")
def test_seg_em():

    # Create a node object
    seg_em = EM()

    # Check if the command is properly defined
    cmd = get_custom_path("seg_EM", env_dir="NIFTYSEGDIR")
    assert seg_em.cmd == cmd

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        seg_em.run()

    # Assign some input data
    in_file = example_data("im1.nii")
    seg_em.inputs.in_file = in_file
    seg_em.inputs.no_prior = 4

    cmd_tmp = "{cmd} -in {in_file} -nopriors 4 -bc_out {bc_out} -out \
{out_file} -out_outlier {out_outlier}"

    expected_cmd = cmd_tmp.format(
        cmd=cmd,
        in_file=in_file,
        out_file="im1_em.nii.gz",
        bc_out="im1_bc_em.nii.gz",
        out_outlier="im1_outlier_em.nii.gz",
    )

    assert seg_em.cmdline == expected_cmd



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The maths module provides higher-level interfaces to some of the operations
that can be performed with the fslmaths command-line program.
"""
import os
import numpy as np

from ..base import TraitedSpec, File, traits, InputMultiPath, isdefined
from .base import FSLCommand, FSLCommandInputSpec


class MathsInput(FSLCommandInputSpec):

    in_file = File(
        position=2, argstr="%s", exists=True, mandatory=True, desc="image to operate on"
    )
    out_file = File(
        genfile=True, position=-2, argstr="%s", desc="image to write", hash_files=False
    )
    _dtypes = ["float", "char", "int", "short", "double", "input"]
    internal_datatype = traits.Enum(
        *_dtypes,
        position=1,
        argstr="-dt %s",
        desc=("datatype to use for calculations " "(default is float)")
    )
    output_datatype = traits.Enum(
        *_dtypes,
        position=-1,
        argstr="-odt %s",
        desc=("datatype to use for output (default " "uses input type)")
    )

    nan2zeros = traits.Bool(
        position=3, argstr="-nan", desc="change NaNs to zeros before doing anything"
    )


class MathsOutput(TraitedSpec):

    out_file = File(exists=True, desc="image written after calculations")


class MathsCommand(FSLCommand):

    _cmd = "fslmaths"
    input_spec = MathsInput
    output_spec = MathsOutput
    _suffix = "_maths"

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_file
        if not isdefined(self.inputs.out_file):
            outputs["out_file"] = self._gen_fname(
                self.inputs.in_file, suffix=self._suffix
            )
        outputs["out_file"] = os.path.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()["out_file"]
        return None


class ChangeDataTypeInput(MathsInput):

    _dtypes = ["float", "char", "int", "short", "double", "input"]
    output_datatype = traits.Enum(
        *_dtypes, position=-1, argstr="-odt %s", mandatory=True, desc="output data type"
    )


class ChangeDataType(MathsCommand):
    """Use fslmaths to change the datatype of an image."""

    input_spec = ChangeDataTypeInput
    _suffix = "_chdt"


class ThresholdInputSpec(MathsInput):

    thresh = traits.Float(
        mandatory=True, position=4, argstr="%s", desc="threshold value"
    )
    direction = traits.Enum(
        "below",
        "above",
        usedefault=True,
        desc="zero-out either below or above thresh value",
    )
    use_robust_range = traits.Bool(
        desc="interpret thresh as percentage (0-100) of robust range"
    )
    use_nonzero_voxels = traits.Bool(
        desc="use nonzero voxels to calculate robust range",
        requires=["use_robust_range"],
    )


class Threshold(MathsCommand):
    """Use fslmaths to apply a threshold to an image in a variety of ways."""

    input_spec = ThresholdInputSpec
    _suffix = "_thresh"

    def _format_arg(self, name, spec, value):
        if name == "thresh":
            arg = "-"
            _si = self.inputs
            if self.inputs.direction == "above":
                arg += "u"
            arg += "thr"
            if isdefined(_si.use_robust_range) and _si.use_robust_range:
                if isdefined(_si.use_nonzero_voxels) and _si.use_nonzero_voxels:
                    arg += "P"
                else:
                    arg += "p"
            arg += " %.10f" % value
            return arg
        return super(Threshold, self)._format_arg(name, spec, value)


class StdImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%sstd",
        position=4,
        desc="dimension to standard deviate across",
    )


class StdImage(MathsCommand):
    """Use fslmaths to generate a standard deviation in an image across a given
    dimension.
    """

    input_spec = StdImageInput
    _suffix = "_std"


class MeanImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%smean",
        position=4,
        desc="dimension to mean across",
    )


class MeanImage(MathsCommand):
    """Use fslmaths to generate a mean image across a given dimension."""

    input_spec = MeanImageInput
    _suffix = "_mean"


class MaxImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%smax",
        position=4,
        desc="dimension to max across",
    )


class MaxImage(MathsCommand):
    """Use fslmaths to generate a max image across a given dimension.

    Examples
    --------
    >>> from nipype.interfaces.fsl.maths import MaxImage
    >>> maxer = MaxImage()
    >>> maxer.inputs.in_file = "functional.nii"  # doctest: +SKIP
    >>> maxer.dimension = "T"
    >>> maxer.cmdline  # doctest: +SKIP
    'fslmaths functional.nii -Tmax functional_max.nii'

    """

    input_spec = MaxImageInput
    _suffix = "_max"


class PercentileImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%sperc",
        position=4,
        desc="dimension to percentile across",
    )
    perc = traits.Range(
        low=0,
        high=100,
        argstr="%f",
        position=5,
        desc=("nth percentile (0-100) of FULL RANGE " "across dimension"),
    )


class PercentileImage(MathsCommand):
    """Use fslmaths to generate a percentile image across a given dimension.

    Examples
    --------
    >>> from nipype.interfaces.fsl.maths import MaxImage
    >>> percer = PercentileImage()
    >>> percer.inputs.in_file = "functional.nii"  # doctest: +SKIP
    >>> percer.dimension = "T"
    >>> percer.perc = 90
    >>> percer.cmdline  # doctest: +SKIP
    'fslmaths functional.nii -Tperc 90 functional_perc.nii'

    """

    input_spec = PercentileImageInput
    _suffix = "_perc"


class MaxnImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%smaxn",
        position=4,
        desc="dimension to index max across",
    )


class MaxnImage(MathsCommand):
    """Use fslmaths to generate an image of index of max across
    a given dimension.

    """

    input_spec = MaxnImageInput
    _suffix = "_maxn"


class MinImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%smin",
        position=4,
        desc="dimension to min across",
    )


class MinImage(MathsCommand):
    """Use fslmaths to generate a minimum image across a given dimension."""

    input_spec = MinImageInput
    _suffix = "_min"


class MedianImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%smedian",
        position=4,
        desc="dimension to median across",
    )


class MedianImage(MathsCommand):
    """Use fslmaths to generate a median image across a given dimension."""

    input_spec = MedianImageInput
    _suffix = "_median"


class AR1ImageInput(MathsInput):

    dimension = traits.Enum(
        "T",
        "X",
        "Y",
        "Z",
        usedefault=True,
        argstr="-%sar1",
        position=4,
        desc=("dimension to find AR(1) coefficient" "across"),
    )


class AR1Image(MathsCommand):
    """Use fslmaths to generate an AR1 coefficient image across a
    given dimension. (Should use -odt float and probably demean first)

    """

    input_spec = AR1ImageInput
    _suffix = "_ar1"


class IsotropicSmoothInput(MathsInput):

    fwhm = traits.Float(
        mandatory=True,
        xor=["sigma"],
        position=4,
        argstr="-s %.5f",
        desc="fwhm of smoothing kernel [mm]",
    )
    sigma = traits.Float(
        mandatory=True,
        xor=["fwhm"],
        position=4,
        argstr="-s %.5f",
        desc="sigma of smoothing kernel [mm]",
    )


class IsotropicSmooth(MathsCommand):
    """Use fslmaths to spatially smooth an image with a gaussian kernel."""

    input_spec = IsotropicSmoothInput
    _suffix = "_smooth"

    def _format_arg(self, name, spec, value):
        if name == "fwhm":
            sigma = float(value) / np.sqrt(8 * np.log(2))
            return spec.argstr % sigma
        return super(IsotropicSmooth, self)._format_arg(name, spec, value)


class ApplyMaskInput(MathsInput):

    mask_file = File(
        exists=True,
        mandatory=True,
        argstr="-mas %s",
        position=4,
        desc="binary image defining mask space",
    )


class ApplyMask(MathsCommand):
    """Use fslmaths to apply a binary mask to another image."""

    input_spec = ApplyMaskInput
    _suffix = "_masked"


class KernelInput(MathsInput):

    kernel_shape = traits.Enum(
        "3D",
        "2D",
        "box",
        "boxv",
        "gauss",
        "sphere",
        "file",
        argstr="-kernel %s",
        position=4,
        desc="kernel shape to use",
    )
    kernel_size = traits.Float(
        argstr="%.4f",
        position=5,
        xor=["kernel_file"],
        desc=(
            "kernel size - voxels for box/boxv, mm " "for sphere, mm sigma for gauss"
        ),
    )
    kernel_file = File(
        exists=True,
        argstr="%s",
        position=5,
        xor=["kernel_size"],
        desc="use external file for kernel",
    )


class DilateInput(KernelInput):

    operation = traits.Enum(
        "mean",
        "modal",
        "max",
        argstr="-dil%s",
        position=6,
        mandatory=True,
        desc="filtering operation to perfoem in dilation",
    )


class DilateImage(MathsCommand):
    """Use fslmaths to perform a spatial dilation of an image."""

    input_spec = DilateInput
    _suffix = "_dil"

    def _format_arg(self, name, spec, value):
        if name == "operation":
            return spec.argstr % dict(mean="M", modal="D", max="F")[value]
        return super(DilateImage, self)._format_arg(name, spec, value)


class ErodeInput(KernelInput):

    minimum_filter = traits.Bool(
        argstr="%s",
        position=6,
        usedefault=True,
        default_value=False,
        desc=("if true, minimum filter rather than " "erosion by zeroing-out"),
    )


class ErodeImage(MathsCommand):
    """Use fslmaths to perform a spatial erosion of an image."""

    input_spec = ErodeInput
    _suffix = "_ero"

    def _format_arg(self, name, spec, value):
        if name == "minimum_filter":
            if value:
                return "-eroF"
            return "-ero"
        return super(ErodeImage, self)._format_arg(name, spec, value)


class SpatialFilterInput(KernelInput):

    operation = traits.Enum(
        "mean",
        "median",
        "meanu",
        argstr="-f%s",
        position=6,
        mandatory=True,
        desc="operation to filter with",
    )


class SpatialFilter(MathsCommand):
    """Use fslmaths to spatially filter an image."""

    input_spec = SpatialFilterInput
    _suffix = "_filt"


class UnaryMathsInput(MathsInput):

    operation = traits.Enum(
        "exp",
        "log",
        "sin",
        "cos",
        "tan",
        "asin",
        "acos",
        "atan",
        "sqr",
        "sqrt",
        "recip",
        "abs",
        "bin",
        "binv",
        "fillh",
        "fillh26",
        "index",
        "edge",
        "nan",
        "nanm",
        "rand",
        "randn",
        "range",
        argstr="-%s",
        position=4,
        mandatory=True,
        desc="operation to perform",
    )


class UnaryMaths(MathsCommand):
    """Use fslmaths to perorm a variety of mathematical operations on an image."""

    input_spec = UnaryMathsInput

    def _list_outputs(self):
        self._suffix = "_" + self.inputs.operation
        return super(UnaryMaths, self)._list_outputs()


class BinaryMathsInput(MathsInput):

    operation = traits.Enum(
        "add",
        "sub",
        "mul",
        "div",
        "rem",
        "max",
        "min",
        mandatory=True,
        argstr="-%s",
        position=4,
        desc="operation to perform",
    )
    operand_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=5,
        xor=["operand_value"],
        desc="second image to perform operation with",
    )
    operand_value = traits.Float(
        argstr="%.8f",
        mandatory=True,
        position=5,
        xor=["operand_file"],
        desc="value to perform operation with",
    )


class BinaryMaths(MathsCommand):
    """Use fslmaths to perform mathematical operations using a second image or
    a numeric value.

    """

    input_spec = BinaryMathsInput


class MultiImageMathsInput(MathsInput):

    op_string = traits.String(
        position=4,
        argstr="%s",
        mandatory=True,
        desc=("python formatted string of operations " "to perform"),
    )
    operand_files = InputMultiPath(
        File(exists=True),
        mandatory=True,
        desc=("list of file names to plug into op " "string"),
    )


class MultiImageMaths(MathsCommand):
    """Use fslmaths to perform a sequence of mathematical operations.

    Examples
    --------
    >>> from nipype.interfaces.fsl import MultiImageMaths
    >>> maths = MultiImageMaths()
    >>> maths.inputs.in_file = "functional.nii"
    >>> maths.inputs.op_string = "-add %s -mul -1 -div %s"
    >>> maths.inputs.operand_files = ["functional2.nii", "functional3.nii"]
    >>> maths.inputs.out_file = "functional4.nii"
    >>> maths.cmdline
    'fslmaths functional.nii -add functional2.nii -mul -1 -div functional3.nii functional4.nii'

    """

    input_spec = MultiImageMathsInput

    def _format_arg(self, name, spec, value):
        if name == "op_string":
            return value % tuple(self.inputs.operand_files)
        return super(MultiImageMaths, self)._format_arg(name, spec, value)


class TemporalFilterInput(MathsInput):

    lowpass_sigma = traits.Float(
        -1,
        argstr="%.6f",
        position=5,
        usedefault=True,
        desc="lowpass filter sigma (in volumes)",
    )
    highpass_sigma = traits.Float(
        -1,
        argstr="-bptf %.6f",
        position=4,
        usedefault=True,
        desc="highpass filter sigma (in volumes)",
    )


class TemporalFilter(MathsCommand):
    """Use fslmaths to apply a low, high, or bandpass temporal filter to a
    timeseries.

    """

    input_spec = TemporalFilterInput
    _suffix = "_filt"
